# -*- coding: utf-8 -*-
from collections import defaultdict
from itertools import product, combinations
from typing import Any, Dict, List, Tuple, Union

from glycopeptidepy.structure.modification.modification import Modification

from glycopeptidepy.utils.collectiontools import descending_combination_counter, _AccumulatorBag
from glycopeptidepy.structure import constants as structure_constants
from glycopeptidepy.structure.composition import Composition
from glycopeptidepy.structure.modification import ModificationCategory, GlycanFragment, ModificationIndex
from glycopeptidepy.structure.glycan import GlycosylationManager, GlycosylationType
from glycopeptidepy.structure.fragment import (
    IonSeries,
    SimpleFragment,
    PeptideFragment,
    format_negative_composition,
    ChemicalShift,
)
from glycopeptidepy.structure.residue import AminoAcidResidue as R, residue_to_neutral_loss as _residue_to_neutral_loss


from .base import (
    FragmentationStrategyBase,
    glycosylation_type_to_core,
    _n_glycosylation,
    _o_glycosylation,
    _gag_linker_glycosylation,
    _modification_xylose,
    _modification_hexnac,
)

from .glycan import _GlycanFragmentingStrategyBase, StubGlycopeptideStrategy


ammonia_loss = ChemicalShift("-NH3", -Composition({"N": 1, "H": 3}))
water_loss = ChemicalShift("-H2O", -Composition({"O": 1, "H": 2}))


class PeptideFragmentationStrategyBase(FragmentationStrategyBase):
    def __init__(
        self, peptide, series, chemical_shifts=None, max_chemical_shifts=1, include_neutral_losses=False, **kwargs
    ):
        if chemical_shifts is None:
            chemical_shifts = dict()
        super(PeptideFragmentationStrategyBase, self).__init__(peptide, **kwargs)
        self.series = IonSeries(series)
        self.chemical_shift_rules = defaultdict(list, chemical_shifts)
        self.max_chemical_shifts = max_chemical_shifts
        self.include_neutral_losses = include_neutral_losses
        self._initialize_fields()

    def _initialize_fields(self):
        self.direction = self.series.direction

        # Null Values
        self.running_mass = 0
        if self.compute_compositions:
            self.running_composition = Composition()
        else:
            self.running_composition = None
        self.index = None
        self.size = len(self.peptide)
        self.modification_index = ModificationIndex()
        self.glycosylation_manager = GlycosylationManager(self.peptide)
        self.amino_acids_counter = _AccumulatorBag()

        self.running_mass += self.series.mass_shift
        if self.compute_compositions:
            self.running_composition += self.series.composition_shift
        self._initialize_start_terminal()

    def _get_viable_chemical_shift_combinations(self):
        shifts = []
        if self.include_neutral_losses:
            shifts.append(ammonia_loss)
            shifts.append(water_loss)
        if not self.chemical_shift_rules:
            return shifts
        for residue, count in self.amino_acids_counter.items():
            loss_composition = self.chemical_shift_rules[residue]
            if loss_composition:
                for i in range(count):
                    shifts.append(loss_composition + [Composition()])
        loss_composition_combinations = {}
        for comb in combinations(shifts, self.max_chemical_shifts):
            for prod in product(*comb):
                loss_composition = sum(prod, Composition())
                loss_composition_combinations[format_negative_composition(loss_composition)] = loss_composition
        return [ChemicalShift(k, v) for k, v in loss_composition_combinations.items() if v]

    def _initialize_start_terminal(self):
        if self.direction > 0:
            self.running_mass += self.peptide.n_term.mass
            if self.compute_compositions:
                self.running_composition += self.peptide.n_term.composition
            self.index = -1
        elif self.direction < 0:
            self.running_mass += self.peptide.c_term.mass
            if self.compute_compositions:
                self.running_composition += self.peptide.c_term.composition
            self.index = len(self.peptide)
        else:
            raise ValueError("Unknown direction %r" % (self.series.direction,))

    def composition_of(self, residue, modifications):
        composition = Composition(residue.composition)
        for mod in modifications:
            composition += mod.composition
        return composition

    def has_more(self):
        if self.direction > 0:
            return self.index < self.size - 2
        else:
            return self.index > 1

    def flanking_residues(self):
        residues = [self.peptide[self.index][0], self.peptide[self.index + self.direction][0]]
        if self.direction < 0:
            residues = residues[::-1]
        return residues

    def name_index_of(self):
        if self.direction > 0:
            return self.index + structure_constants.FRAG_OFFSET
        else:
            return (self.size - self.index - 1) + structure_constants.FRAG_OFFSET

    def __next__(self):
        if self.has_more():
            return self.step()
        else:
            raise StopIteration()

    def track_glycosylation(self, index, glycosylation):
        self.glycosylation_manager[self.index] = glycosylation
        self.modification_index[glycosylation] += 1

    def _update_state(self):
        self.index += self.direction
        residue, modifications = self.peptide[self.index]
        for mod in modifications:
            if mod.is_tracked_for(ModificationCategory.glycosylation):
                self.track_glycosylation(self.index, mod)
            else:
                self.modification_index[mod] += 1
        self.amino_acids_counter[residue] += 1
        self.running_mass += residue.mass
        if self.compute_compositions:
            composition = self.composition_of(residue, modifications)
            self.running_composition += composition

    def _build_fragments(self):
        fragments_from_site = []
        frag = PeptideFragment(
            self.series,
            self.name_index_of(),
            self.modification_index.copy(),
            self.running_mass,
            glycosylation=self.glycosylation_manager.copy(),
            flanking_amino_acids=self.flanking_residues(),
            composition=self.running_composition,
        )
        shifts = self._get_viable_chemical_shift_combinations()
        for fragment in self.partial_loss(frag):
            fragments_from_site.append(fragment)
            for shift in shifts:
                f = fragment.clone()
                f.chemical_shift = shift
                fragments_from_site.append(f)
        return fragments_from_site

    def step(self):
        self._update_state()
        fragments_from_site = self._build_fragments()
        return fragments_from_site

    def __repr__(self):
        return "%s(%s, %r, %r, %0.4f, %d)" % (
            self.__class__.__name__,
            self.peptide,
            self.series,
            self.running_composition,
            self.running_mass,
            self.index,
        )


class HCDFragmentationStrategy(PeptideFragmentationStrategyBase):
    modifications_of_interest = {
        k.name: k
        for k in (
            [_n_glycosylation, _modification_hexnac, _o_glycosylation, _gag_linker_glycosylation, _modification_xylose]
        )
    }

    modification_compositions = {k.name: k.composition for k in modifications_of_interest.values()}

    def __init__(
        self, peptide, series, chemical_shifts=None, max_chemical_shifts=1, include_neutral_losses=False, **kwargs
    ):
        super(HCDFragmentationStrategy, self).__init__(
            peptide, series, chemical_shifts, max_chemical_shifts, include_neutral_losses, **kwargs
        )
        self._last_modification_set = None
        self._last_modification_variants = None

    def _get_core_for(self, glycosylation):
        try:
            if not glycosylation.rule.is_core:
                glycosylation = glycosylation_type_to_core[glycosylation.rule.glycosylation_type]()
            return glycosylation
        except KeyError:
            raise ValueError("Cannot determine which core to use for {}".format(glycosylation.rule.glycosylation_type))

    def composition_of(self, residue, modifications):
        composition = Composition(residue.composition)
        for mod in modifications:
            if mod.is_tracked_for(ModificationCategory.glycosylation):
                mod = self._get_core_for(mod)
            composition += mod.composition
        return composition

    def track_glycosylation(self, index, glycosylation):
        # HCD strategy does not track intact topologies
        try:
            glycosylation = self._get_core_for(glycosylation)
        except KeyError:
            raise ValueError("Cannot determine which core to use for {}".format(glycosylation.rule.glycosylation_type))
        self.glycosylation_manager[self.index] = glycosylation
        self.modification_index[glycosylation] += 1

    def _get_modifications_of_interest(self, fragment):
        modifications = fragment.modification_dict
        if self.compute_compositions:
            delta_composition = Composition()
        else:
            delta_composition = None
        other_modifications = ModificationIndex()
        modifications_of_interest = ModificationIndex()

        for k, v in modifications.items():
            if k.name in self.modifications_of_interest:
                modifications_of_interest[k] = v
                if self.compute_compositions:
                    delta_composition += self.modification_compositions[k.name] * v
            else:
                other_modifications[k] = v

        return modifications_of_interest, delta_composition, other_modifications

    def _replace_cores(self, modifications_of_interest):
        n_cores = modifications_of_interest.pop(_n_glycosylation, 0)
        o_cores = modifications_of_interest.pop(_o_glycosylation, 0)
        gag_cores = modifications_of_interest.pop(_gag_linker_glycosylation, 0)

        # Convert core glycosylation into the residual monosaccharides remaining
        # after dissociation
        hexnac_cores = n_cores * 2 + o_cores
        if hexnac_cores:
            modifications_of_interest[_modification_hexnac] += hexnac_cores
        if gag_cores:
            modifications_of_interest[_modification_xylose] += gag_cores
        return modifications_of_interest

    def _generate_modification_variants(self, interesting_modifications, other_modifications):
        variants = []
        for varied_modifications in descending_combination_counter(interesting_modifications):
            updated_modifications = ModificationIndex(other_modifications)
            for k, v in varied_modifications.items():
                if v != 0:
                    updated_modifications[k] += v
            if self.compute_compositions:
                extra_composition = Composition()
                for mod, mod_count in varied_modifications.items():
                    extra_composition += self.modification_compositions[mod.name] * mod_count
            variants.append((updated_modifications, extra_composition))
        return variants

    def partial_loss(self, fragment):
        (modifications_of_interest, delta_composition, other_modifications) = self._get_modifications_of_interest(
            fragment
        )

        if self.compute_compositions:
            base_composition = fragment.composition - delta_composition
        else:
            base_composition = None

        self._replace_cores(modifications_of_interest)

        if (modifications_of_interest, other_modifications) == self._last_modification_set:
            variants = self._last_modification_variants
        else:
            self._last_modification_set = modifications_of_interest, other_modifications
            variants = self._last_modification_variants = self._generate_modification_variants(
                modifications_of_interest, other_modifications
            )

        series = self.series
        fragments = []
        for updated_modifications, extra_composition in variants:
            fragments.append(
                PeptideFragment(
                    series,
                    fragment.position,
                    updated_modifications,
                    fragment.bare_mass,
                    flanking_amino_acids=fragment.flanking_amino_acids,
                    composition=(base_composition + extra_composition) if self.compute_compositions else None,
                )
            )
        return fragments


try:
    _has_c = True
    from glycopeptidepy._c.structure.fragmentation_strategy.peptide import (
        PeptideFragmentationStrategyBase,
        HCDFragmentationStrategy,
    )
except ImportError:
    _has_c = False

# Cooper, H. J., Hudgins, R. R., Håkansson, K., & Marshall, A. G. (2002).
# Characterization of amino acid side chain losses in electron capture dissociation.
# Journal of the American Society for Mass Spectrometry, 13(3), 241–249.
# https://doi.org/10.1016/S1044-0305(01)00357-9
exd_sidechain_losses = defaultdict(
    list,
    {
        R("His"): [-Composition("H6C4N2")],
        R("Arg"): [-Composition("CH5N3"), -Composition("C4H11N3"), -Composition("C1N2H4"), -Composition("N2H6")],
        R("Asn"): [-Composition("CONH3")],
        R("Gln"): [-Composition("CONH3")],
        R("Met"): [-Composition("C3H6S")],
    },
)

for key, value in _residue_to_neutral_loss.items():
    exd_sidechain_losses[R(key)].extend(value)


class GlycanCompositionFragment(SimpleFragment):
    __slots__ = ("glycan_composition",)

    def __init__(self, name, mass, kind, composition, glycan_composition, chemical_shift=None):
        super(GlycanCompositionFragment, self).__init__(
            name, mass, kind, composition, chemical_shift=chemical_shift, is_glycosylated=True
        )
        self.glycan_composition = glycan_composition

    def clone(self):
        return self.__class__(
            self.name,
            self.mass,
            self.kind,
            self.composition,
            self.glycan_composition,
            self.chemical_shift,
        )


class _GlycanFragmentingMixin(_GlycanFragmentingStrategyBase):
    max_glycan_cleavages: int
    allow_ambiguous_fragmentation: bool
    _glycan_fragment_cache: Dict[Tuple, List]
    _glycan_fragment_wrapper_cache: Dict[Any, GlycanFragment]
    _glycan_fragment_combination_cache: Dict

    def _make_fragmentation_strategy(self):
        strat = StubGlycopeptideStrategy(self.peptide, True)
        return strat

    def _get_glycan_composition_fragments(
        self, glycosylation: Modification, series: str, position: int
    ) -> List[GlycanCompositionFragment]:
        strat = self._make_fragmentation_strategy()
        gc = strat.glycan_composition()
        if glycosylation.rule.glycosylation_type == GlycosylationType.n_linked:
            gen = strat.n_glycan_composition_fragments(gc)
        elif glycosylation.rule.glycosylation_type == GlycosylationType.o_linked:
            gen = strat.o_glycan_composition_fragments(gc)
        elif glycosylation.rule.glycosylation_type == GlycosylationType.glycosaminoglycan:
            gen = strat.gag_linker_composition_fragments(gc)
        else:
            gen = []
        fragments = []
        for frag_spec in gen:
            key = frag_spec["key"]
            if not key:
                continue
            name = "".join("%s%d" % kv for kv in sorted(key.items(), key=lambda x: x[0].mass()))
            mass = frag_spec["mass"]
            composition = frag_spec["composition"]
            fragments.append(GlycanCompositionFragment(name, mass, IonSeries.other, composition, frag_spec["key"]))
        return fragments

    def _get_glycan_structure_fragments(self, glycosylation: Modification, series: str, position: int):
        fragments = sorted(
            glycosylation.rule.get_fragments(series, max_cleavages=self.max_glycan_cleavages), key=lambda x: x.mass
        )
        return fragments

    def _get_glycan_fragments(
        self, glycosylation: Modification, series: str, position: int
    ) -> List[Union[GlycanCompositionFragment, GlycanFragment]]:
        try:
            return self._glycan_fragment_cache[glycosylation, series, position]
        except KeyError:
            if glycosylation.rule.is_composition:
                if glycosylation.rule.is_core and not self.allow_ambiguous_fragmentation:
                    fragments = []
                else:
                    fragments = self._get_glycan_composition_fragments(glycosylation, series, position)
            else:
                fragments = self._get_glycan_structure_fragments(glycosylation, series, position)
            self._glycan_fragment_cache[glycosylation, series, position] = fragments
            return fragments

    def _wrap_glycan_fragment(self, fragment) -> GlycanFragment:
        try:
            return self._glycan_fragment_wrapper_cache[fragment]
        except KeyError:
            result = self._glycan_fragment_wrapper_cache[fragment] = GlycanFragment(fragment)
            return result


class EXDFragmentationStrategy(PeptideFragmentationStrategyBase, _GlycanFragmentingMixin):
    glycan_fragment_ladder = "Y"

    def __init__(
        self,
        peptide,
        series,
        chemical_shifts=None,
        max_chemical_shifts=1,
        max_glycan_cleavages=None,
        include_neutral_losses=False,
        allow_ambiguous_fragmentation=True,
        **kwargs,
    ):
        super(EXDFragmentationStrategy, self).__init__(
            peptide, series, chemical_shifts, max_chemical_shifts, include_neutral_losses, **kwargs
        )
        if max_glycan_cleavages is None:
            max_glycan_cleavages = self._guess_max_glycan_cleavages()
        self.max_glycan_cleavages = max_glycan_cleavages
        self.allow_ambiguous_fragmentation = allow_ambiguous_fragmentation
        self._glycan_fragment_cache = {}
        self._glycan_fragment_wrapper_cache = {}
        self._glycan_fragment_combination_cache = {}

    def _strip_glycosylation_from_modification_dict(self, modification_dict: Dict[int, Modification]):
        glycosylations = []
        stripped_modifications = {}
        for position, glycosylation in sorted(self.glycosylation_manager.items()):
            glycosylations.append((position, glycosylation))
        for modification, count in modification_dict.items():
            if modification.is_tracked_for(ModificationCategory.glycosylation):
                continue
            stripped_modifications[modification] = count
        return glycosylations, stripped_modifications

    def _get_glycan_loss_combinations(self, glycosylations, glycan_fragment_ladder):
        glycosylations = tuple(glycosylations)
        try:
            wrapped_fragment_combinations = self._glycan_fragment_combination_cache[
                glycosylations, glycan_fragment_ladder
            ]
        except KeyError:
            glycan_fragments = [
                [None] + self._get_glycan_fragments(glycosylation, glycan_fragment_ladder, position)
                for position, glycosylation in glycosylations
            ]
            fragment_combinations = product(*glycan_fragments)

            has_multiple_nonlocalized_cores = len(glycosylations) > 1 and sum(
                g[1].rule.is_core for g in glycosylations
            ) == len(glycosylations)

            wrapped_fragment_combinations = []
            for fragment_set in fragment_combinations:
                if fragment_set == ():
                    continue
                aggregate_glycan_composition = None
                if has_multiple_nonlocalized_cores:
                    aggregate_glycan_composition = _AccumulatorBag()
                subfragments = []
                for glycan_fragment in fragment_set:
                    if glycan_fragment is None:
                        continue
                    subfragment = self._wrap_glycan_fragment(glycan_fragment)
                    if has_multiple_nonlocalized_cores and glycan_fragment is not None:
                        aggregate_glycan_composition += glycan_fragment.glycan_composition
                    subfragments.append(subfragment)
                if aggregate_glycan_composition is not None:
                    invalid = False
                    for k, v in self.peptide.glycan_composition.items():
                        if aggregate_glycan_composition[k] > v:
                            invalid = True
                            break
                    if invalid:
                        continue
                wrapped_fragment_combinations.append(subfragments)
            self._glycan_fragment_combination_cache[
                glycosylations, glycan_fragment_ladder
            ] = wrapped_fragment_combinations
        return wrapped_fragment_combinations

    def partial_loss(self, fragment: PeptideFragment, glycan_fragment_ladder=None):
        if glycan_fragment_ladder is None:
            glycan_fragment_ladder = self.glycan_fragment_ladder
        # make a clone in case we mutate anything
        base = fragment.clone()

        (glycosylations, stripped_modifications) = self._strip_glycosylation_from_modification_dict(
            base.modification_dict
        )
        if self.compute_compositions:
            base_composition = base.composition
            for _glycosylation_position, glycosylation in glycosylations:
                base_composition -= glycosylation.composition
        else:
            base_composition = None
        bare_fragment = PeptideFragment(
            fragment.series,
            fragment.position,
            ModificationIndex(stripped_modifications),
            fragment.bare_mass,
            flanking_amino_acids=fragment.flanking_amino_acids,
            composition=base_composition.clone() if self.compute_compositions else None,
        )

        wrapped_fragment_combinations = self._get_glycan_loss_combinations(glycosylations, glycan_fragment_ladder)

        results = []
        for fragment_set in wrapped_fragment_combinations:
            if self.compute_compositions:
                delta_composition = Composition()
            else:
                delta_composition = None
            new_modifications = ModificationIndex(stripped_modifications)
            for subfragment in fragment_set:
                new_modifications[subfragment] += 1

            extended_fragment = PeptideFragment(
                bare_fragment.series,
                bare_fragment.position,
                new_modifications,
                bare_fragment.bare_mass,
                flanking_amino_acids=bare_fragment.flanking_amino_acids,
                composition=bare_fragment.composition + delta_composition if self.compute_compositions else None,
            )
            results.append(extended_fragment)
        results.append(fragment)
        return results


class CachingEXDFragmentationStrategy(EXDFragmentationStrategy):
    shared_composition_fragment_cache_ambiguous: Dict[str, List[GlycanCompositionFragment]] = defaultdict(dict)
    shared_composition_fragment_combination_cache_ambiguous: Dict[str, List[GlycanCompositionFragment]] = defaultdict(dict)

    def __init__(
        self,
        peptide,
        series,
        chemical_shifts=None,
        max_chemical_shifts=1,
        max_glycan_cleavages=None,
        include_neutral_losses=False,
        allow_ambiguous_fragmentation=True,
        **kwargs,
    ):
        super().__init__(
            peptide,
            series,
            chemical_shifts,
            max_chemical_shifts,
            max_glycan_cleavages,
            include_neutral_losses,
            allow_ambiguous_fragmentation,
            **kwargs,
        )

        self.is_all_core = self._infer_shared_cache_keys()

    def _infer_shared_cache_keys(self):
        manager: GlycosylationManager = self.peptide.glycosylation_manager
        is_composition = not manager.is_fully_specified_topologies()
        num_glycans = len(manager)

        if is_composition:
            if num_glycans == 1:
                key = str(manager.aggregate)
                cache = self.shared_composition_fragment_cache_ambiguous[key]

                if cache:
                    position, marker = next(iter(manager.items()))
                    series = "Y"

                    self._glycan_fragment_cache[(marker, series, position)] = cache

                combo_cache = self.shared_composition_fragment_combination_cache_ambiguous[key]
                if combo_cache:
                    self._glycan_fragment_combination_cache[(((position, marker),), series)] = combo_cache[0]
                return True

            else:
                glycosylation_sites = list(manager.items())
                all_cores = all(event.rule.is_core for _site, event in  glycosylation_sites)

                if all_cores:
                    series = "Y"
                    key = str(manager.aggregate)
                    glycosylation_sites.sort(key=lambda x: x[0])

                    cache = self.shared_composition_fragment_cache_ambiguous[key]
                    if cache:
                        for (position, marker) in glycosylation_sites:
                            self._glycan_fragment_cache[(marker, series, position)] = cache

                    combo_cache = self.shared_composition_fragment_combination_cache_ambiguous[key]
                    if combo_cache:
                        combo_cache.sort(key=len)

                        for i in range(1, len(glycosylation_sites) + 1):
                            combos = tuple(glycosylation_sites[:i])
                            self._glycan_fragment_combination_cache[(combos, series)] = combo_cache[:i]

                return all_cores

    def _update_shared_cache_keys(self):
        if self.is_all_core:
            manager: GlycosylationManager = self.peptide.glycosylation_manager
            key = str(manager.aggregate)
            if self._glycan_fragment_cache:
                self.shared_composition_fragment_cache_ambiguous[key] = next(iter(self._glycan_fragment_cache.values()))

            if self._glycan_fragment_combination_cache:
                combinations = list(self._glycan_fragment_combination_cache.items())
                combinations.sort(key=lambda x: len(x[0][0]))
                self.shared_composition_fragment_combination_cache_ambiguous[key] = [stack[1] for stack in combinations]

    def _get_glycan_loss_combinations(self, glycosylations, glycan_fragment_ladder):
        size_before = len(self._glycan_fragment_combination_cache)
        result = super()._get_glycan_loss_combinations(glycosylations, glycan_fragment_ladder)
        size_after = len(self._glycan_fragment_combination_cache)

        if size_before != size_after:
            self._update_shared_cache_keys()

        return result


from glycopeptidepy.structure import (PeptideSequence, AminoAcidResidue,
                                      Modification, ModificationNameResolutionError, AnonymousModificationRule)


class MzIdentMLPeptideParser(object):
    def __init__(self, modification_translation_table=None):
        if modification_translation_table is None:
            modification_translation_table = {}
        self.modification_translation_table = modification_translation_table

    def process(self, peptide_dict):
        peptide_sequence = PeptideSequence(peptide_dict["PeptideSequence"])
        peptide_sequence.id = peptide_dict.get('id')
        self.handle_substitutions(peptide_dict, peptide_sequence)
        self.handle_modifications(peptide_dict, peptide_sequence)
        return peptide_sequence

    def __call__(self, peptide_dict):
        return self.process(peptide_dict)

    def handle_substitutions(self, peptide_dict, peptide_sequence):
        if "SubstitutionModification" in peptide_dict:
            subs = peptide_dict["SubstitutionModification"]
            for sub in subs:
                pos = sub['location'] - 1
                replace = AminoAcidResidue(sub["replacementResidue"])
                peptide_sequence.substitute(pos, replace)

    def add_modification(self, peptide_sequence, modification, position):
        if position == -1:
            targets = modification.rule.n_term_targets
            for t in targets:
                if t.position_modifier is not None and t.amino_acid_targets is None:
                    break
            else:
                position += 1
        if position == len(peptide_sequence):
            targets = modification.rule.c_term_targets
            for t in targets:
                if t.position_modifier is not None and t.amino_acid_targets is None:
                    break
            else:
                position -= 1

        if position == -1:
            peptide_sequence.n_term = modification
        elif position == len(peptide_sequence):
            peptide_sequence.c_term = modification
        else:
            peptide_sequence.add_modification(position, modification)

    def handle_modifications(self, peptide_dict, peptide_sequence):
        if "Modification" in peptide_dict:
            mods = peptide_dict["Modification"]
            for mod in mods:
                pos = mod["location"] - 1
                accession = None
                try:
                    if "unknown modification" in mod:
                        try:
                            _name = mod['unknown modification']
                            if _name in self.modification_translation_table:
                                modification = self.modification_translation_table[_name](
                                )
                            else:
                                modification = Modification(str(_name))
                        except ModificationNameResolutionError:
                            raise KeyError("Cannot find key in %r" % (mod,))
                    else:
                        try:
                            _name = mod["name"]
                            accession = getattr(_name, "accession", None)
                            _name = str(_name)
                            if accession is not None:
                                accession = str(accession)
                                try:
                                    modification = Modification(accession)
                                except ModificationNameResolutionError:
                                    modification = Modification(_name)
                            else:
                                modification = Modification(_name)

                        except (KeyError, ModificationNameResolutionError) as e:
                            raise KeyError(
                                "Cannot find key %s in %r" % (e, mod))

                    self.add_modification(peptide_sequence, modification, pos)
                except KeyError:
                    if "unknown modification" in mod:
                        if 'monoisotopicMassDelta' in mod:
                            mass = float(mod['monoisotopicMassDelta'])
                            modification = AnonymousModificationRule(
                                str(_name), mass)()
                            self.add_modification(
                                peptide_sequence, modification, pos)
                        else:
                            raise
                    else:
                        raise


parse = MzIdentMLPeptideParser()



from .structure.sequence import PeptideSequence, ProteinSequence
from .structure.modification import (
    Modification, ModificationTable, RestrictedModificationTable,
    ModificationRule, ModificationTarget, AnonymousModificationRule,
    Glycosylation, SequenceLocation, ModificationCategory)
from .structure.fragment import (
    IonSeries, PeptideFragment, ChemicalShift,
    SimpleFragment, StubFragment)
from .structure.residue import (
    AminoAcidResidue, get_all_residues, register_degenerate, register_residue)
from .enzyme import Protease, cleave
from .structure.glycan import (
    HashableGlycanComposition, GlycosylationType, TypedGlycanComposition)
from .structure.composition import (Composition, ChemicalCompositionError)


def parse(string, peptide_class=PeptideSequence):
    return peptide_class(string)


__all__ = [
    "PeptideSequence", "ProteinSequence", "parse",

    "AminoAcidResidue", "get_all_residues", "register_residue", "register_degenerate",

    "Modification", "ModificationTable", "RestrictedModificationTable",
    "ModificationRule", "ModificationTarget", "AnonymousModificationRule",

    "IonSeries", "PeptideFragment", "ChemicalShift", "SimpleFragment", "StubFragment",

    "Protease", "cleave",

    "HashableGlycanComposition", "GlycosylationType", "TypedGlycanComposition",
    "Glycosylation",

    "Composition", "ChemicalCompositionError",
]



from typing import List, Optional
from glypy.utils import make_struct


class MoleculeBase(object):
    __slots__ = ()
    mass = None

    def __copy__(self):
        return self.clone()


# A few base types for doing type-based behavior changes
class PeptideSequenceBase(MoleculeBase):
    '''
    A base type for classes describing peptide sequences, with or without modifiations
    '''

    def _invalidate(self):
        pass


class ModificationBase(MoleculeBase):
    '''
    A base type for classes describing peptide sequence modifications
    '''
    __slots__ = ()

    name: str

    def serialize(self):
        '''A string representation for inclusion in sequences'''
        return self.name


class ResidueBase(MoleculeBase):
    '''
    A base type for classes describing amino acid residues
    '''
    __slots__ = ()


class SequencePosition(make_struct('SequencePosition', ['amino_acid', 'modifications'])):
    __slots__ = ()

    amino_acid: ResidueBase
    modifications: Optional[List[ModificationBase]]

    def __init__(self, parts):
        super(SequencePosition, self).__init__(*parts)

    def __repr__(self):
        return repr(list(self))

    @property
    def mass(self):
        mass = self.amino_acid.mass
        for mod in self.modifications:
            mass += mod.mass
        return mass


try:
    from glycopeptidepy._c.structure.base import (
        AminoAcidResidueBase as ResidueBase,
        ModificationBase, PeptideSequenceBase,
        SequencePosition)
except ImportError:
    pass


__all__ = [
    "ResidueBase", "ModificationBase", "PeptideSequenceBase",
    "SequencePosition"
]



try:
    from glycopeptidepy._c.structure.constants import Configuration
    constants = Configuration()
except ImportError:
    from argparse import Namespace
    constants = Namespace()

# Tokenizer Constants
constants.N_TERM_DEFAULT = "H"
constants.C_TERM_DEFAULT = "OH"

# N-glycan Sequon detection
constants.ALLOW_MODIFIED_ASPARAGINE = False

# Sequence Fragment Constants
constants.FRAG_OFFSET = 1
constants.PARTIAL_HEXNAC_LOSS = True
constants.EXCLUDE_B1 = True



# cython: embedsignature=True
cimport cython

from cpython.ref cimport Py_INCREF
from cpython cimport PyObject
from cpython.float cimport PyFloat_AsDouble
from cpython.tuple cimport PyTuple_GetItem, PyTuple_Size, PyTuple_GET_ITEM, PyTuple_GET_SIZE
from cpython.list cimport PyList_GetItem, PyList_SetItem, PyList_Size, PyList_New, PyList_GET_ITEM, PyList_GET_SIZE, PyList_Append, PyList_SET_ITEM
from cpython.int cimport PyInt_AsLong
from cpython.dict cimport (PyDict_GetItem, PyDict_SetItem, PyDict_Next,
                           PyDict_Keys, PyDict_Update, PyDict_DelItem, PyDict_Size)

from glypy.composition.ccomposition cimport CComposition
from glypy._c.structure.glycan_composition cimport _CompositionBase

from glycopeptidepy._c.structure.fragmentation_strategy.base cimport FragmentationStrategyBase

from glycopeptidepy._c.count_table cimport CountTable, CountTableIterator
from glycopeptidepy._c.structure.fragment cimport (
    IonSeriesBase, StubFragment, _NameTree,
    build_name_from_composition)

from glycopeptidepy._c.structure.sequence_methods cimport _PeptideSequenceCore
from glycopeptidepy._c.structure.glycan cimport GlycosylationManager, GlycanCompositionProxy, glycan_composition_type

from glycopeptidepy.structure.fragment import (
    ChemicalShift, IonSeries, format_negative_composition)

from glycopeptidepy.structure.glycan import (
    GlycosylationType,
    HashableGlycanComposition)

from itertools import product as _product
from glypy.structure.glycan_composition import (
    FrozenMonosaccharideResidue)

cdef object product = _product

cdef:
    object GlycosylationType_n_linked = GlycosylationType.n_linked
    object GlycosylationType_o_linked = GlycosylationType.o_linked
    object GlycosylationType_glycosaminoglycan = GlycosylationType.glycosaminoglycan

cdef IonSeriesBase IonSeries_stub_glycopeptide = IonSeries.stub_glycopeptide


@cython.final
cdef class _CompositionTree(object):
    cdef:
        public _NameTree root

    def __init__(self, root=None):
        if root is None:
            root = _NameTree()
        self.root = root

    def __getitem__(self, key):
        return self.root[key]

    cpdef _NameTree build(self, list pair_sequence):
        cdef:
            tuple item
            int i, n
            _CompositionBase value
            _NameTree node
        node = self.root.traverse(pair_sequence)
        if node.name is None:
            # modify the hashable glycan composition before calculating
            # its hash value
            value = node.name = <_CompositionBase>HashableGlycanComposition()
            n = PyList_Size(pair_sequence)
            for i in range(n):
                item = <tuple>PyList_GET_ITEM(pair_sequence, i)
                k = <object>PyTuple_GET_ITEM(item, 0)
                v = <object>PyTuple_GET_ITEM(item, 1)
                value._setitem_fast(k, v)
            # Force the population of the _mass and _str caches
            node.name.mass()
            node.value = str(node.name)
        return node


cdef _CompositionTree _composition_tree_root = _CompositionTree()
cdef dict _composition_name_cache = dict()


cdef _NameTree _prepare_glycan_composition_from_mapping(CountTable mapping):
    pair_sequence = mapping.items()
    return _composition_tree_root.build(pair_sequence)

cdef:
    object _HEX = FrozenMonosaccharideResidue.from_iupac_lite("Hex")
    CComposition _HEX_composition = _HEX.total_composition()
    double _HEX_mass = _HEX.mass()
    object _HEXNAC = FrozenMonosaccharideResidue.from_iupac_lite("HexNAc")
    CComposition _HEXNAC_composition = _HEXNAC.total_composition()
    double _HEXNAC_mass = _HEXNAC.mass()
    object _FUC = FrozenMonosaccharideResidue.from_iupac_lite("Fuc")
    CComposition _FUC_composition = _FUC.total_composition()
    double _FUC_mass = _FUC.mass()
    object _DHEX = FrozenMonosaccharideResidue.from_iupac_lite("dHex")
    CComposition _DHEX_composition = _DHEX.total_composition()
    double _DHEX_mass = _DHEX.mass()
    object _XYL = FrozenMonosaccharideResidue.from_iupac_lite("Xyl")
    CComposition _XYL_composition = _XYL.total_composition()
    double _XYL_mass = _XYL.mass()
    object _AHEX = FrozenMonosaccharideResidue.from_iupac_lite("aHex")
    CComposition _AHEX_composition = _AHEX.total_composition()
    double _AHEX_mass = _AHEX.mass()

    object _Hex2NAc = FrozenMonosaccharideResidue.from_iupac_lite("Hex2NAc")
    object _Glc2NAc = FrozenMonosaccharideResidue.from_iupac_lite("Glc2NAc")
    object _Gal2NAc = FrozenMonosaccharideResidue.from_iupac_lite("Gal2NAc")


cdef class GlycanCompositionFragmentStrategyBase(FragmentationStrategyBase):
    def __init__(self, peptide, use_query=False, *args, **kwargs):
        super(GlycanCompositionFragmentStrategyBase, self).__init__(peptide, *args, **kwargs)
        # to be initialized closer to when the glycan composition will
        # be used. Determine whether to use the fast-path __getitem__ or
        # go through the slower but more general GlycanComposition.query
        # method to look up monosaccharides.
        self._use_query = use_query
        self._generator = None
        if self.peptide is not None:
            self.glycosylation_manager = self.peptide._glycosylation_manager
        else:
            self.glycosylation_manager = None

    cpdef GlycanCompositionProxy glycan_composition(self):
        return self.glycosylation_manager.get_glycan_composition()

    cdef bint _guess_query_mode(self, GlycanCompositionProxy glycan_composition):
        # these guesses will work for N-glycans and common types of mucin-type O-glycans
        # and GAG linkers
        flag = PyInt_AsLong(glycan_composition._getitem_fast(_Hex2NAc)) +\
               PyInt_AsLong(glycan_composition._getitem_fast(_Glc2NAc)) +\
               PyInt_AsLong(glycan_composition._getitem_fast(_Gal2NAc))
        return flag or self._use_query

    cpdef long count_glycosylation_type(self, glycotype):
        return GlycosylationManager.count_glycosylation_type(self.glycosylation_manager, glycotype)


@cython.freelist(10000)
cdef class GlycanCompositionFragment(object):

    def __init__(self, mass, composition, key, is_extended=False):
        self.mass = mass
        self.composition = composition
        self.key = key
        self.is_extended = is_extended
        self._hash_key = -1
        self._glycosylation_size = -1
        self.cache = None

    def __hash__(self):
        cdef _NameTree node
        if self._hash_key == -1:
            node = self._get_glycan_composition()
            self._hash_key = hash(node.name)
        return self._hash_key

    cpdef _NameTree _get_glycan_composition(self):
        return _prepare_glycan_composition_from_mapping(self.key)

    def __eq__(self, other):
        return self.key == other.key and self.is_extended == other.is_extended

    @staticmethod
    cdef GlycanCompositionFragment _create(double mass, CComposition composition, CountTable key, bint is_extended):
        cdef:
            GlycanCompositionFragment self
        self = GlycanCompositionFragment.__new__(GlycanCompositionFragment)
        self.mass = mass
        self.composition = composition
        self.key = key
        self.is_extended = is_extended
        self._hash_key = -1
        self.cache = None
        return self

    def __getitem__(self, key):
        if key == "mass":
            return self.mass
        elif key == "composition":
            return self.composition
        elif key == "key":
            return self.key
        elif key == 'is_extended':
            return self.is_extended
        else:
            raise KeyError(key)

    def __setitem__(self, key, value):
        if key == "mass":
            self.mass = value
        elif key == "composition":
            self.composition = value
        elif key == "key":
            self.key = value
        elif key == 'is_extended':
            self.is_extended = value
        else:
            raise KeyError(key)

    def __repr__(self):
        template = "{self.__class__.__name__}({self.mass}, {self.composition}, {self.key}, {self.is_extended})"
        return template.format(self=self)

    cpdef GlycanCompositionFragment copy(self):
        cdef:
            GlycanCompositionFragment inst
        inst = GlycanCompositionFragment._create(
            self.mass,
            self.composition.clone() if self.composition is not None else None, self.key.copy(),
            self.is_extended)
        inst._hash_key = self._hash_key
        inst.cache = self.cache
        return inst

    cdef int get_glycosylation_size(self):
        if self._glycosylation_size == -1:
            self._glycosylation_size = self.key.total()
        return self._glycosylation_size

    @property
    def glycosylation_size(self):
        return self.get_glycosylation_size()


cdef dict hexnac_hex_composition_cache = dict()

cdef CComposition hexnac_hex_composition(long hexnac, long hexose):
    cdef:
        tuple key
        CComposition result
        PyObject* ptemp

    key = (hexnac, hexose)
    ptemp = PyDict_GetItem(hexnac_hex_composition_cache, key)
    if ptemp == NULL:
        result = (_HEXNAC_composition * hexnac)
        if hexose > 0:
            result.add_from(<CComposition>(_HEX_composition * hexose))
        PyDict_SetItem(hexnac_hex_composition_cache, key, result.clone())
        return result
    else:
        result = <CComposition>ptemp
        return result.clone()


cdef class StubGlycopeptideStrategy(GlycanCompositionFragmentStrategyBase):

    """A fragmentation strategy that generates intact peptide + glycan Y fragments from
    glycopeptides where the glycan structure is not fully known. When the structure is
    known, :class:`CADFragmentationStrategy` should be used instead

    Attributes
    ----------
    extended : bool
        Whether to fragment beyond the conserved core
    extended_fucosylation : bool
        Whether to consider multiple ``Fuc`` residues per fragment rather than one per core
    """

    _composition_tree = _composition_tree_root

    def __init__(self, peptide, extended=True, use_query=False, extended_fucosylation=False, **kwargs):
        self.extended = extended
        self.extended_fucosylation = extended_fucosylation
        super(StubGlycopeptideStrategy, self).__init__(peptide, use_query)

    def __next__(self):
        if self._generator is None:
            self._generator = iter(self.stub_fragments())
        return next(self._generator)

    cpdef bint _validate_glycan_composition(self, CountTable aggregate_glycosylation, glycan):
        cdef:
            bint invalid = False
            int i, n
            long v, ov
            object key
            PyObject* pkey
            CountTableIterator iterator
        if self._use_query:
            for key, value in aggregate_glycosylation.items():
                if glycan.query(key) < value:
                    invalid = True
                    break
        else:
            iterator = CountTableIterator._create(aggregate_glycosylation)
            pkey = NULL
            getter = glycan._getitem_fast
            while iterator.has_more():
                iterator.get_next_value(&pkey, &v)
                key = <object>pkey
                ov = PyInt_AsLong(getter(key))
                if ov < v:
                    invalid = True
                    break
        return invalid

    cpdef GlycanCompositionFragment fucosylate_increment(self, GlycanCompositionFragment shift):
        cdef:
            GlycanCompositionFragment fucosylated
        fucosylated = shift.copy()
        fucosylated.mass += _FUC_mass
        if self.compute_compositions:
            fucosylated.composition.add_from(_FUC_composition)
        fucosylated.key.increment(_FUC, 1)
        return fucosylated

    cpdef GlycanCompositionFragment xylosylate_increment(self, GlycanCompositionFragment shift):
        cdef:
            GlycanCompositionFragment xylosylated
        xylosylated = shift.copy()
        xylosylated.mass += _XYL_mass
        if self.compute_compositions:
            xylosylated.composition.add_from(_XYL_composition)
        xylosylated.key.increment(_XYL, 1)
        return xylosylated

    cpdef list fucosylate_extended(self, GlycanCompositionFragment shift, long fucose_count):
        cdef:
            GlycanCompositionFragment fucosylated
            size_t i
            list result
        result = [None for i in range(fucose_count)]
        for i in range(1, fucose_count + 1):
            fucosylated = shift.copy()
            fucosylated.mass += _FUC_mass * i
            if self.compute_compositions:
                fucosylated.composition.add_from(<CComposition>(i * _FUC_composition))
            fucosylated.key.increment(_FUC, i)
            result[i - 1] = fucosylated
        return result

    cpdef GlycanCompositionFragment modified_increment(self, GlycanCompositionFragment base, object shift):
        cdef:
            GlycanCompositionFragment modified
        modified = base.copy()

        modified.mass += shift.mass()
        if self.compute_compositions:
            modified.composition.add_from(shift.total_composition())
        modified.key.increment(shift, 1)
        return modified

    cpdef list n_glycan_composition_fragments(self, object glycan, int core_count=1, int  iteration_count=0):
        """Generate theoretical N-glycan Y fragment compositions containing the core motif
        plus a portion of the extended branches if :attr:`extended` is used. Unless :attr:`extend

        Parameters
        ----------
        glycan : :class:`~.GlycanComposition`
            Description
        core_count : int, optional
            The total number of glycans attached to the peptide
        iteration_count : int, optional
            The core index. Unless :attr:`extended_fucosylation` is true, this will
            limit the number of Fucose residues to one per core at most, and if
            the iteration count exceeds the number of cores no Fucose will be added

        Returns
        -------
        list of dict
            The compositions corresponding to the theoretical fragments
        """

        cdef:
            long fucose_count, xylose_count
            long hexnac_count, hexose_count
            long hexnac_in_aggregate, hexose_in_aggregate
            long base_hexnac, extra_hexnac_count , extra_hexose_count
            CountTable key_ct
            list core_shifts

        if self._use_query:
            fucose_count = glycan.query('Fuc') + glycan.query('dHex')
            xylose_count = glycan.query('Xyl')
            hexnac_in_aggregate = glycan.query('HexNAc')
            hexose_in_aggregate = glycan.query('Hex')
        else:
            fucose_count = glycan[_FUC] + glycan[_DHEX]
            xylose_count = glycan[_XYL]
            hexnac_in_aggregate = glycan[_HEXNAC]
            hexose_in_aggregate = glycan[_HEX]

        core_shifts = []
        base_hexnac = min(hexnac_in_aggregate + 1, 3)
        for hexnac_count in range(base_hexnac):
            if hexnac_count == 0:
                shift = GlycanCompositionFragment._create(0,
                    CComposition() if self.compute_compositions else None,
                    CountTable._create(),
                    False)
                core_shifts.append(shift)
            elif hexnac_count == 1:
                key_ct = CountTable._create()
                key_ct.setitem(_HEXNAC, hexnac_count)
                shift = GlycanCompositionFragment._create(
                    (hexnac_count * _HEXNAC_mass),
                    hexnac_hex_composition(hexnac_count, 0) if self.compute_compositions else None,
                    key_ct,
                    False,
                )
                core_shifts.append(shift)
                if iteration_count < fucose_count:
                    fucosylated = self.fucosylate_increment(shift)
                    core_shifts.append(fucosylated)
            elif hexnac_count == 2:
                key_ct = CountTable._create()
                key_ct.setitem(_HEXNAC, hexnac_count)
                shift = GlycanCompositionFragment._create(
                    (hexnac_count * _HEXNAC_mass),
                    hexnac_hex_composition(hexnac_count, 0) if self.compute_compositions else None,
                    key_ct,
                    False,
                )
                core_shifts.append(shift)

                if not self.extended_fucosylation:
                    if iteration_count < fucose_count:
                        fucosylated = self.fucosylate_increment(shift)
                        core_shifts.append(fucosylated)
                        if iteration_count < xylose_count:
                            xylosylated = self.xylosylate_increment(fucosylated)
                            core_shifts.append(xylosylated)
                elif fucose_count > 0:
                    core_shifts.extend(self.fucosylate_extended(shift, fucose_count))
                    if iteration_count < xylose_count:
                        xylosylated = self.xylosylate_increment(fucosylated)
                        core_shifts.append(xylosylated)

                if iteration_count < xylose_count:
                    xylosylated = self.xylosylate_increment(shift)
                    core_shifts.append(xylosylated)

                for hexose_count in range(1, min(hexose_in_aggregate + 1, 4)):
                    key_ct = CountTable._create()
                    key_ct.setitem(_HEXNAC, hexnac_count)
                    key_ct.setitem(_HEX, hexose_count)
                    shift = GlycanCompositionFragment._create(
                        (hexnac_count * _HEXNAC_mass) + (hexose_count * _HEX_mass),
                        hexnac_hex_composition(hexnac_count, hexose_count) if self.compute_compositions else None,
                        key_ct,
                        False,
                    )
                    core_shifts.append(shift)

                    if not self.extended_fucosylation:
                        if iteration_count < fucose_count:
                            fucosylated = self.fucosylate_increment(shift)
                            core_shifts.append(fucosylated)
                            if iteration_count < xylose_count:
                                xylosylated = self.xylosylate_increment(fucosylated)
                                core_shifts.append(xylosylated)
                    elif fucose_count > 0:
                        core_shifts.extend(self.fucosylate_extended(shift, fucose_count))
                        if iteration_count < xylose_count:
                            xylosylated = self.xylosylate_increment(fucosylated)
                            core_shifts.append(xylosylated)

                    if iteration_count < xylose_count:
                        xylosylated = self.xylosylate_increment(shift)
                        core_shifts.append(xylosylated)

                    # After the core motif has been exhausted, speculatively add
                    # on the remaining core monosaccharides sequentially until
                    # exhausted. This doesn't allow for one missing alpha-mannose while still
                    # showing an extension on the other.
                    if hexose_count == 3 and hexnac_in_aggregate >= 2 * core_count and self.extended:
                        for extra_hexnac_count in range(0, hexnac_in_aggregate - hexnac_count + 1):
                            if extra_hexnac_count + hexnac_count > hexnac_in_aggregate:
                                continue
                            if extra_hexnac_count > 0:
                                key_ct = CountTable._create()
                                key_ct.setitem(_HEXNAC, hexnac_count + extra_hexnac_count)
                                key_ct.setitem(_HEX, hexose_count)
                                shift = GlycanCompositionFragment._create(
                                    ((hexnac_count + extra_hexnac_count) * _HEXNAC_mass) + (
                                        hexose_count * _HEX_mass),
                                    hexnac_hex_composition(hexnac_count + extra_hexnac_count, hexose_count) if self.compute_compositions else None,
                                    key_ct,
                                    True
                                )
                                core_shifts.append(shift)

                                if not self.extended_fucosylation:
                                    if iteration_count < fucose_count:
                                        fucosylated = self.fucosylate_increment(shift)
                                        core_shifts.append(fucosylated)
                                        if iteration_count < xylose_count:
                                            xylosylated = self.xylosylate_increment(fucosylated)
                                            core_shifts.append(xylosylated)
                                elif fucose_count > 0:
                                    core_shifts.extend(self.fucosylate_extended(shift, fucose_count))

                                if iteration_count < xylose_count:
                                    xylosylated = self.xylosylate_increment(shift)
                                    core_shifts.append(xylosylated)

                            for extra_hexose_count in range(1, hexose_in_aggregate - hexose_count + 1):
                                if extra_hexose_count + hexose_count > hexose_in_aggregate:
                                    continue
                                key_ct = CountTable._create()
                                key_ct.setitem(_HEXNAC, hexnac_count + extra_hexnac_count)
                                key_ct.setitem(_HEX, hexose_count + extra_hexose_count)
                                shift = GlycanCompositionFragment._create(
                                    ((hexnac_count + extra_hexnac_count) * _HEXNAC_mass) + (
                                        (hexose_count + extra_hexose_count) * _HEX_mass),
                                    hexnac_hex_composition(
                                        hexnac_count + extra_hexnac_count, hexose_count + extra_hexose_count) if self.compute_compositions else None,

                                    key_ct,
                                    True
                                )
                                core_shifts.append(shift)

                                if not self.extended_fucosylation:
                                    if iteration_count < fucose_count:
                                        fucosylated = self.fucosylate_increment(shift)
                                        core_shifts.append(fucosylated)
                                        if iteration_count < xylose_count:
                                            xylosylated = self.xylosylate_increment(fucosylated)
                                            core_shifts.append(xylosylated)
                                elif fucose_count > 0:
                                    core_shifts.extend(self.fucosylate_extended(shift, fucose_count))
                                    if iteration_count < xylose_count:
                                        xylosylated = self.xylosylate_increment(fucosylated)
                                        core_shifts.append(xylosylated)

                                if iteration_count < xylose_count:
                                    xylosylated = self.xylosylate_increment(shift)
                                    core_shifts.append(xylosylated)
        return core_shifts

    cdef list _combinate_sites_single(self, list per_site_shifts, object glycan):
        cdef:
            list result
            Py_ssize_t i, n
            GlycanCompositionFragment intermediate
            CountTable aggregate_glycosylation
            _NameTree glycosylation
            PyObject* ptemp
            tuple triple

        result = []
        n = PyList_GET_SIZE(per_site_shifts)
        for i in range(n):
            intermediate = <GlycanCompositionFragment>PyList_GET_ITEM(per_site_shifts, i)
            aggregate_glycosylation = intermediate.key
            glycosylation = intermediate.cache
            if glycosylation is None:
                glycosylation = _prepare_glycan_composition_from_mapping(
                    aggregate_glycosylation)
                intermediate.cache = glycosylation
            name_key = glycosylation.value

            ptemp = PyDict_GetItem(_composition_name_cache, name_key)
            if ptemp == NULL:
                name = build_name_from_composition(aggregate_glycosylation)
                PyDict_SetItem(_composition_name_cache, name_key, name)
            else:
                name = <str>ptemp
            triple = (intermediate, name, glycosylation.name)
            PyList_Append(result, triple)
        return result


    cpdef list _combinate_sites(self, list per_site_shifts, object glycan):
        '''Combine the glycan fragments from each site together to form an aggregate
        fragment glycan composition.
        '''
        cdef:
            list result
            list combos
            Py_ssize_t i, n, j, m
            tuple positions
            bint is_extended, invalid
            CComposition composition
            GlycanCompositionFragment intermediate, site
            CountTable aggregate_glycosylation
            _NameTree glycosylation
            set seen
            int core_count
            double total_mass
            double acc_mass

        core_count = PyList_Size(per_site_shifts)
        if core_count == 1:
            # Fast path that doesn't need to use itertools.product or any of the extra
            # unpacking and validation
            combos = <list>PyList_GET_ITEM(per_site_shifts, 0)
            return self._combinate_sites_single(combos, glycan)
        combos = list(product(*per_site_shifts))
        result = []
        seen = set()
        n = PyList_Size(combos)
        total_mass = PyFloat_AsDouble(glycan.mass())
        for i in range(n):
            positions = <tuple>PyList_GET_ITEM(combos, i)
            aggregate_glycosylation = None
            glycosylation = None
            composition = None

            intermediate = <GlycanCompositionFragment>PyTuple_GET_ITEM(positions, 0)
            acc_mass = intermediate.mass
            m = core_count

            # A quick pre-scan to skip all combinations that exceed the mass of the total
            for j in range(1, m):
                site = <GlycanCompositionFragment>PyTuple_GET_ITEM(positions, j)
                acc_mass += site.mass
            invalid = total_mass < acc_mass
            if invalid:
                continue

            # Now do the expensive accumulation
            aggregate_glycosylation = intermediate.key.copy()
            if self.compute_compositions:
                composition = intermediate.composition.copy()
            else:
                composition = None
            is_extended = intermediate.is_extended
            for j in range(1, m):
                site = <GlycanCompositionFragment>PyTuple_GET_ITEM(positions, j)
                aggregate_glycosylation._add_from(site.key)
                if self.compute_compositions:
                    composition.add_from(site.composition)
                is_extended |= site.is_extended
            invalid = self._validate_glycan_composition(
                aggregate_glycosylation, glycan)
            if invalid:
                continue

            intermediate = GlycanCompositionFragment._create(
                acc_mass, composition, aggregate_glycosylation, is_extended)

            if glycosylation is None:
                glycosylation = _prepare_glycan_composition_from_mapping(
                    aggregate_glycosylation)
                if core_count == 1:
                    intermediate.cache = glycosylation
            name_key = glycosylation.value
            # May not need to do this second round of building.
            ptemp = PyDict_GetItem(_composition_name_cache, name_key)
            if ptemp == NULL:
                name = build_name_from_composition(aggregate_glycosylation)
                PyDict_SetItem(_composition_name_cache, name_key, name)
            else:
                name = <str>ptemp
            if name in seen:
                continue
            seen.add(name)
            result.append((intermediate, name, glycosylation.name))
        return result

    cpdef list n_glycan_stub_fragments(self):
        cdef:
            int core_count
            list per_site_shifts
            GlycanCompositionFragment site
            CComposition base_composition, composition
            double base_mass
            size_t i
            size_t n_combos
            bint is_extended, is_glycosylated
            str name_key, name
            object glycosylation
            GlycanCompositionProxy glycan
            list accumulator
            list combos
            tuple result

        glycan = self.glycan_composition()
        self._use_query = self._guess_query_mode(glycan)
        core_count = self.count_glycosylation_type(GlycosylationType_n_linked)
        per_site_shifts = []
        base_composition = None
        if self.compute_compositions:
            base_composition = self.peptide_composition()
            base_mass = base_composition.calc_mass()
        else:
            base_mass = self.peptide.get_peptide_backbone_mass()
        for i in range(core_count):
            core_shifts = self.n_glycan_composition_fragments(glycan, core_count, i)
            per_site_shifts.append(core_shifts)
        combos = self._combinate_sites(per_site_shifts, glycan)
        n_combos = PyList_Size(combos)
        accumulator = PyList_New(n_combos)
        for i in range(n_combos):
            result = <tuple>PyList_GetItem(combos, i)
            site = <GlycanCompositionFragment>PyTuple_GetItem(result, 0)
            name = <str>PyTuple_GetItem(result, 1)
            glycosylation = <object>PyTuple_GetItem(result, 2)
            is_glycosylated = (site.mass != 0)

            composition = None
            if self.compute_compositions:
                composition = base_composition.clone()
                composition.add_from(site.composition)

            frag = StubFragment._create(
                name=name,
                mass=base_mass + site.mass,
                composition=composition,
                is_glycosylated=is_glycosylated,
                kind=IonSeries_stub_glycopeptide,
                chemical_shift=None,
                glycosylation=glycosylation,
                is_extended=site.is_extended,
                glycosylation_size=site.get_glycosylation_size())
            Py_INCREF(frag)
            PyList_SET_ITEM(accumulator, i, frag)
        return accumulator

    cpdef list o_glycan_stub_fragments(self):
        cdef:
            list per_site_shifts
            CComposition base_composition, composition
            double base_mass, mass
            size_t core_count, i, n_positions
            tuple positions
            GlycanCompositionFragment site
            CountTable aggregate_glycosylation
            bint is_glycosylated, is_extended
            GlycanCompositionProxy glycan
            str name
            set seen
            list accumulator

        glycan = self.glycan_composition()
        self._use_query = self._guess_query_mode(glycan)
        core_count = self.count_glycosylation_type(GlycosylationType_o_linked)
        per_site_shifts = []

        base_composition = None
        if self.compute_compositions:
            base_composition = self.peptide_composition()
            base_mass = base_composition.calc_mass()
        else:
            base_mass = self.peptide.get_peptide_backbone_mass()

        for i in range(core_count):
            core_shifts = self.o_glycan_composition_fragments(glycan, core_count, i)
            per_site_shifts.append(core_shifts)
        combos = self._combinate_sites(per_site_shifts, glycan)
        n_combos = PyList_Size(combos)
        accumulator = PyList_New(n_combos)
        for i in range(n_combos):
            result = <tuple>PyList_GetItem(combos, i)
            site = <GlycanCompositionFragment>PyTuple_GetItem(result, 0)
            name = <str>PyTuple_GetItem(result, 1)
            glycosylation = <object>PyTuple_GetItem(result, 2)
            is_glycosylated = (site.mass != 0)
            composition = None
            if self.compute_compositions:
                composition = base_composition.clone()
                composition.add_from(site.composition)

            frag = StubFragment._create(
                name=name,
                mass=base_mass + site.mass,
                composition=composition,
                is_glycosylated=is_glycosylated,
                kind=IonSeries_stub_glycopeptide,
                chemical_shift=None,
                glycosylation=glycosylation,
                is_extended=site.is_extended,
                glycosylation_size=site.get_glycosylation_size())
            Py_INCREF(frag)
            PyList_SET_ITEM(accumulator, i, frag)
        return accumulator

    cpdef list o_glycan_composition_fragments(self, glycan, long core_count=1, long iteration_count=0):
        cdef:
            list core_shifts
            long fucose_count, hexnac_in_aggregate, hexose_in_aggregate
            long hexnac_count, hexose_count
            long extra_hexnac_count, extra_hexose_count
            CountTable key_ct

        if self._use_query:
            fucose_count = glycan.query('Fuc') + glycan.query('dHex')
            hexnac_in_aggregate = glycan.query('HexNAc')
            hexose_in_aggregate = glycan.query('Hex')
        else:
            fucose_count = glycan[_FUC] + glycan[_DHEX]
            hexnac_in_aggregate = glycan[_HEXNAC]
            hexose_in_aggregate = glycan[_HEX]
        core_shifts = []
        for hexnac_count in range(3):
            if hexnac_in_aggregate < hexnac_count:
                continue
            if hexnac_count == 0:
                shift = GlycanCompositionFragment._create(
                    0,
                    CComposition(),
                    CountTable._create(),
                    False
                )
                core_shifts.append(shift)
            elif hexnac_count >= 1:
                key_ct = CountTable._create()
                key_ct.setitem(_HEXNAC, hexnac_count)
                shift = GlycanCompositionFragment._create(
                    (hexnac_count * _HEXNAC_mass),
                    hexnac_hex_composition(hexnac_count, 0),
                    key_ct,
                    False
                )
                core_shifts.append(shift)
                if iteration_count < fucose_count:
                    fucosylated = self.fucosylate_increment(shift)
                    core_shifts.append(fucosylated)
                for hexose_count in range(0, 2):
                    if hexose_in_aggregate < hexose_count:
                        continue
                    if hexose_count > 0:
                        key_ct = CountTable._create()
                        key_ct.setitem(_HEXNAC, hexnac_count)
                        key_ct.setitem(_HEX, hexose_count)
                        shift = GlycanCompositionFragment._create(
                            (
                                (hexnac_count) * _HEXNAC_mass) + (
                                (hexose_count) * _HEX_mass),
                            hexnac_hex_composition(hexnac_count, hexose_count),
                            key_ct,
                            False
                        )
                        core_shifts.append(shift)
                        if iteration_count < fucose_count:
                            fucosylated = self.fucosylate_increment(shift)
                            core_shifts.append(fucosylated)
                    # After the core motif has been exhausted, speculatively add
                    # on the remaining core monosaccharides sequentially until
                    # exhausted.
                    if self.extended and hexnac_in_aggregate - hexnac_count >= 0:
                        for extra_hexnac_count in range(hexnac_in_aggregate - hexnac_count + 1):
                            if extra_hexnac_count:
                                key_ct = CountTable._create()
                                key_ct.setitem(_HEXNAC, hexnac_count + extra_hexnac_count)
                                if hexose_count > 0:
                                    key_ct.setitem(_HEX, hexose_count)
                                shift = GlycanCompositionFragment._create(
                                    (
                                        (hexnac_count + extra_hexnac_count) * _HEXNAC_mass) + (
                                        (hexose_count) * _HEX_mass),
                                    hexnac_hex_composition(hexnac_count + extra_hexnac_count, hexose_count),
                                    key_ct,
                                    True,
                                )
                                core_shifts.append(shift)
                                if iteration_count < fucose_count:
                                    fucosylated = self.fucosylate_increment(shift)
                                    core_shifts.append(fucosylated)
                            if hexose_in_aggregate > hexose_count and hexose_count > 0:
                                for extra_hexose_count in range(hexose_in_aggregate - hexose_count):
                                    if extra_hexose_count > 0 and extra_hexose_count + hexose_count > 0:
                                        key_ct = CountTable._create()
                                        if hexnac_count > 0:
                                            key_ct.setitem(_HEXNAC, hexnac_count + extra_hexnac_count)
                                        key_ct.setitem(_HEX, hexose_count + extra_hexose_count)
                                        shift = GlycanCompositionFragment._create(
                                            (
                                                (hexnac_count + extra_hexnac_count) * _HEXNAC_mass) + (
                                                (hexose_count + extra_hexose_count) * _HEX_mass),
                                            hexnac_hex_composition(
                                                hexnac_count + extra_hexnac_count, hexose_count + extra_hexose_count),
                                            key_ct,
                                            True
                                        )
                                        core_shifts.append(shift)
                                        if iteration_count < fucose_count:
                                            fucosylated = self.fucosylate_increment(
                                                shift)
                                            core_shifts.append(fucosylated)

        return core_shifts

    cpdef list gag_linker_stub_fragments(self):
        cdef:
            list per_site_shifts
            CComposition base_composition, composition
            double base_mass, mass
            size_t core_count, i, n_positions
            tuple positions
            GlycanCompositionFragment site
            CountTable aggregate_glycosylation
            bint is_glycosylated, is_extended
            GlycanCompositionProxy glycan
            str name
            set seen
            list accumulator

        glycan = self.glycan_composition()
        self._use_query = self._guess_query_mode(glycan)
        core_count = self.count_glycosylation_type(GlycosylationType_glycosaminoglycan)
        per_site_shifts = []
        base_composition = None
        if self.compute_compositions:
            base_composition = self.peptide_composition()
            base_mass = base_composition.calc_mass()
        else:
            base_mass = self.peptide.get_peptide_backbone_mass()
        for i in range(core_count):
            core_shifts = self.gag_linker_composition_fragments(glycan, core_count, i)
            per_site_shifts.append(core_shifts)
        combos = self._combinate_sites(per_site_shifts, glycan)
        n_combos = PyList_Size(combos)
        accumulator = []
        for i in range(n_combos):
            result = <tuple>PyList_GetItem(combos, i)
            site = <GlycanCompositionFragment>PyTuple_GetItem(result, 0)
            name = <str>PyTuple_GetItem(result, 1)
            glycosylation = <object>PyTuple_GetItem(result, 2)
            is_glycosylated = (site.mass != 0)
            composition = None
            if self.compute_compositions:
                composition = base_composition.clone()
                composition.add_from(site.composition)
            accumulator.append(
                StubFragment._create(
                    name=name,
                    mass=base_mass + site.mass,
                    composition=composition,
                    is_glycosylated=is_glycosylated,
                    kind=IonSeries_stub_glycopeptide,
                    chemical_shift=None,
                    glycosylation=glycosylation,
                    is_extended=site.is_extended,
                    glycosylation_size=site.get_glycosylation_size()))
        return accumulator

    cpdef list gag_linker_composition_fragments(self, glycan, long core_count=1, long iteration_count=0):
        cdef:
            long xyl_in_aggregate
            long xyl_count, hexose_count
            list core_shifts
            CountTable key_ct
        xyl_in_aggregate = glycan[_XYL]
        if xyl_in_aggregate == 0:
            xyl_in_aggregate = glycan.query("Pen", exact=False)
        core_shifts = []
        for xyl_count in range(max(0, xyl_in_aggregate) + 1):
            if xyl_count == 0:
                shift = GlycanCompositionFragment._create(
                    0,
                    CComposition(),
                    CountTable._create(),
                    False
                )
                core_shifts.append(shift)
            else:
                key_ct = CountTable._create()
                key_ct.setitem(_XYL, xyl_count)
                shift = GlycanCompositionFragment._create(
                    _XYL_mass * xyl_count,
                    _XYL_composition * xyl_count,
                    key_ct,
                    False
                )
                core_shifts.append(shift)
            if xyl_count > 0:
                # TODO: Handle modified Hexose residues here too.
                for hexose_count in range(1, 3):
                    key_ct = CountTable._create()
                    key_ct.setitem(_XYL, xyl_count)
                    key_ct.setitem(_HEX, hexose_count)
                    shift = GlycanCompositionFragment._create(
                        ((_XYL_mass * xyl_count) + (
                            _HEX_mass * hexose_count)),
                        (
                            (_XYL_composition * xyl_count) + (
                                _HEX_composition * hexose_count)),
                        key_ct,
                        False
                    )
                    core_shifts.append(shift)
                    if hexose_count == 2:
                        key_ct = CountTable._create()
                        key_ct.setitem(_XYL, xyl_count)
                        key_ct.setitem(_HEX, hexose_count)
                        key_ct.setitem(_AHEX, 1)
                        shift = GlycanCompositionFragment._create(
                            ((_XYL_mass * xyl_count) + (
                                _HEX_mass * hexose_count) + _AHEX_mass),
                            (
                                (_XYL_composition * xyl_count) + (
                                    _HEX_composition * hexose_count) + _AHEX_composition),
                            key_ct,
                            False
                        )
                        core_shifts.append(shift)
        return core_shifts

    cpdef list mixed_stub_fragments(self):

        cdef:
            int core_count
            int total_core_count
            list per_site_shifts
            GlycanCompositionFragment site
            CComposition base_composition, composition
            double base_mass
            size_t i
            size_t n_combos
            bint is_extended, is_glycosylated
            str name_key, name
            object glycosylation
            GlycanCompositionProxy glycan
            list accumulator
            list combos
            tuple result

        total_core_count = 0
        glycan = self.glycan_composition()
        self._use_query = self._guess_query_mode(glycan)
        per_site_shifts = []
        base_composition = None
        if self.compute_compositions:
            base_composition = self.peptide_composition()
            base_mass = base_composition.calc_mass()
        else:
            base_mass = self.peptide.get_peptide_backbone_mass()

        core_count = self.count_glycosylation_type(GlycosylationType_n_linked)
        for i in range(core_count):
            core_shifts = self.n_glycan_composition_fragments(glycan, core_count, i)
            per_site_shifts.append(core_shifts)

        total_core_count += core_count
        core_count = self.count_glycosylation_type(GlycosylationType_o_linked)
        for i in range(core_count):
            core_shifts = self.o_glycan_composition_fragments(glycan, core_count, i)
            per_site_shifts.append(core_shifts)

        total_core_count += core_count
        core_count = self.count_glycosylation_type(GlycosylationType_glycosaminoglycan)
        for i in range(core_count):
            core_shifts = self.gag_linker_composition_fragments(glycan, core_count, i)
            per_site_shifts.append(core_shifts)

        core_count += total_core_count
        combos = self._combinate_sites(per_site_shifts, glycan)
        n_combos = PyList_Size(combos)
        accumulator = PyList_New(n_combos)
        for i in range(n_combos):
            result = <tuple>PyList_GetItem(combos, i)
            site = <GlycanCompositionFragment>PyTuple_GetItem(result, 0)
            name = <str>PyTuple_GetItem(result, 1)
            glycosylation = <object>PyTuple_GetItem(result, 2)
            is_glycosylated = (site.mass != 0)

            composition = None
            if self.compute_compositions:
                composition = base_composition.clone()
                composition.add_from(site.composition)

            frag = StubFragment._create(
                name=name,
                mass=base_mass + site.mass,
                composition=composition,
                is_glycosylated=is_glycosylated,
                kind=IonSeries_stub_glycopeptide,
                chemical_shift=None,
                glycosylation=glycosylation,
                is_extended=site.is_extended,
                glycosylation_size=site.get_glycosylation_size())
            Py_INCREF(frag)
            PyList_SET_ITEM(accumulator, i, frag)
        return accumulator

    cpdef list stub_fragments(self):
        n_glycan = self.count_glycosylation_type(GlycosylationType_n_linked) > 0
        o_glycan = self.count_glycosylation_type(GlycosylationType_o_linked) > 0
        gag_linker = self.count_glycosylation_type(GlycosylationType_glycosaminoglycan) > 0
        if (n_glycan + o_glycan + gag_linker) > 1:
            return self.mixed_stub_fragments()
        elif n_glycan:
            return self.n_glycan_stub_fragments()
        elif o_glycan:
            return self.o_glycan_stub_fragments()
        elif gag_linker:
            return self.gag_linker_stub_fragments()
        else:
            if self.glycosylation_manager.get_size() > 0:
                raise ValueError("Unknown Glycan Class Detected")
        return []


from .descriptors import (ModificationCategory, SequenceLocation)
from .utils import (ModificationIndex, ModificationNameResolutionError, ModificationStringParseError)

from .target import (
    ModificationTarget, get_position_modifier_rules_dict,
    extract_targets_from_string, extract_targets_from_rule_string,
    title_cleaner, is_mass_delta)

from .rule import (ModificationRule, AnonymousModificationRule, AminoAcidSubstitution)
from .glycosylation import (
    Glycosylation, CoreGlycosylation, NGlycanCoreGlycosylation,
    MucinOGlycanCoreGlycosylation, OGlycanCoreGlycosylation, GlycosaminoglycanLinkerGlycosylation,
    OGlcNAcylation, GlycanFragment, hexnac_modification, xylose_modification,
    parse_glycan, glycan_resolvers, glycoct, iupac, linear_code, wurcs)

from .source import (
    ModificationSource, ModificationTable, RestrictedModificationTable, rule_string_to_specialized_rule)

from .modification import Modification


__all__ = [
    "ModificationCategory", "SequenceLocation",
    "ModificationIndex", "ModificationNameResolutionError", "ModificationStringParseError",

    "ModificationTarget", "get_position_modifier_rules_dict",
    "extract_targets_from_string", "extract_targets_from_rule_string",
    "title_cleaner", "is_mass_delta",

    "ModificationRule", "AnonymousModificationRule", "AminoAcidSubstitution",

    "Glycosylation", "CoreGlycosylation", "NGlycanCoreGlycosylation",
    "MucinOGlycanCoreGlycosylation", "OGlycanCoreGlycosylation", "GlycosaminoglycanLinkerGlycosylation",
    "OGlcNAcylation", "GlycanFragment", "hexnac_modification", "xylose_modification",
    "parse_glycan", "glycan_resolvers", "glycoct", "iupac", "linear_code", "wurcs",

    "ModificationSource", "ModificationTable", "RestrictedModificationTable",
    "rule_string_to_specialized_rule",

    "Modification",
]



import unittest

from glycopeptidepy.structure.modification import Modification, ModificationTarget, SequenceLocation
from glycopeptidepy.structure import sequence
from glycopeptidepy.algorithm import reverse_preserve_sequon, edit_distance, PeptidoformGenerator

p2 = "YPVLN(N-Glycosylation)VTMPN(Deamidation)NGKFDK{Hex:9; HexNAc:2}"


heparanase = ''.join("""MLLRSKPALPPPLMLLLLGPLGPLSPGALPRPAQAQDVVDLDFFT
QEPLHLVSPSFLSVTIDANLATDPRFLILLGSPKLRTLARGLSPA
YLRFGGTKTDFLIFDPKKESTFEERSYWQSQVNQDICKYGSIPPD
VEEKLRLEWPYQEQLLLREHYQKKFKNSTYSRSSVDVLYTFANCS
GLDLIFGLNALLRTADLQWNSSNAQLLLDYCSSKGYNISWELGNE
PNSFLKKADIFINGSQLGEDFIQLHKLLRKSTFKNAKLYGPDVGQ
PRRKTAKMLKSFLKAGGEVIDSVTWHHYYLNGRTATKEDFLNPDV
LDIFISSVQKVFQVVESTRPGKKVWLGETSSAYGGGAPLLSDTFA
AGFMWLDKLGLSARMGIEVVMRQVFFGAGNYHLVDENFDPLPDYW
LSLLFKKLVGTKVLMASVQGSKRRKLRVYLHCTNTDNPRYKEGDL
TLYAINLHNVTKYLRLPYPFSNKQVDKYLLRPLGPHGLLSKSVQL
NGLTLKMVDDQTLPPLMEKPLRPGSSLGLPAFSYSFFVIRNAKVA
ACI""".splitlines())


class TestReversePreserveSequonTarget(unittest.TestCase):
    def test_with_glycan(self):
        ref = sequence.PeptideSequence(p2)
        mass = ref.total_mass
        composition = ref.total_composition()

        perm = reverse_preserve_sequon(ref)
        self.assertAlmostEqual(mass, perm.total_mass, 4)
        self.assertAlmostEqual(ref.total_mass, perm.total_mass, 4)
        self.assertEqual(composition, perm.total_composition())
        self.assertEqual(ref.total_composition(), perm.total_composition())

    def test_detect_similarity(self):
        fwd = sequence.PeptideSequence("PEPEPR")
        rev = reverse_preserve_sequon(fwd)
        assert edit_distance(fwd, rev) > 0


class TestEditDistance(unittest.TestCase):
    def test_distance(self):
        assert edit_distance("PEPTIDE", "PIPTIDE") == 1
        assert edit_distance(sequence.PeptideSequence("PEPTIDE"),
                             sequence.PeptideSequence("PIPTIDE")) == 1

    def test_reverse_distance(self):
        fwd = sequence.PeptideSequence(heparanase)
        rev = reverse_preserve_sequon(fwd)
        assert edit_distance(fwd, rev) > 0


class TestPeptidoformGenerator(unittest.TestCase):
    def test_generate(self):
        Deamidation = Modification("Deamidation").rule
        duplicate = Deamidation.clone()
        duplicate.targets = {ModificationTarget("N")}
        oxidupe = Modification("Oxidation").rule.clone()
        oxidupe.targets = {ModificationTarget(None, SequenceLocation.protein_c_term)}
        generator = PeptidoformGenerator([], [duplicate, oxidupe])
        seq = sequence.PeptideSequence("PEPQSINIDE")
        results = list(generator(seq))
        self.assertEqual(len(results), 2)

        seq = sequence.PeptideSequence("PEPQSINIDE")
        results = list(generator(seq, protein_c_term=True))
        self.assertEqual(len(results), 4)

    def test_throughput(self):
        Deamidation = Modification("Deamidation").rule
        Oxidation = Modification("Oxidation").rule
        generator = PeptidoformGenerator([], [Deamidation, Oxidation])
        seq = sequence.PeptideSequence(heparanase[:24])
        results = list(generator(seq))
        self.assertEqual(len(results), 8898)


if __name__ == '__main__':
    unittest.main()



from collections import defaultdict, Counter
from itertools import product, combinations, combinations_with_replacement
from typing import List
from glycopeptidepy.structure.modification.glycosylation import Glycosylation

from glypy.composition import formula
from glypy.structure.glycan_composition import (
    FrozenMonosaccharideResidue,
    FrozenGlycanComposition,
    MonosaccharideResidue,
    SubstituentResidue,
    HashableGlycanComposition)

from glycopeptidepy.utils.collectiontools import descending_combination_counter, _AccumulatorBag
from glycopeptidepy.structure.composition import (
    Composition)

from glycopeptidepy.structure.glycan import (
    GlycosylationType,)

from glycopeptidepy.structure.fragment import (
    IonSeries,
    OxoniumIon,
    SimpleFragment,
    StubFragment,
    _NameTree,
    NeutralLoss)


from .base import FragmentationStrategyBase


oxonium_ion_series = IonSeries.oxonium_ion

_substituent_residue_cache = {}
labile_substituents = ('sulfate', 'phosphate')


def remove_labile_modifications(residue):
    has_copied = False
    labile_substituents_removed = None
    try:
        for position, substituent in residue.substituents():
            if substituent.name in labile_substituents:
                if labile_substituents_removed is None:
                    labile_substituents_removed = []
                try:
                    substituent_residue = _substituent_residue_cache[substituent.name]
                except KeyError:
                    substituent_residue = _substituent_residue_cache[substituent.name] = SubstituentResidue(substituent.name)
                labile_substituents_removed.append(substituent_residue)
                if not has_copied:
                    residue = residue.clone(
                        monosaccharide_type=MonosaccharideResidue)
                residue.drop_substituent(position, substituent)
    except AttributeError:
        pass
    return residue, labile_substituents_removed


def separate_labile_modifications(glycan_composition):
    labile_modifications = []
    new_counts = HashableGlycanComposition()
    for residue, count in glycan_composition.items():
        new_residue, separated_modifications = remove_labile_modifications(residue)
        new_counts[new_residue] += count
        if separated_modifications:
            for mod in separated_modifications:
                new_counts[mod] += 1
            labile_modifications.extend(labile_modifications)
    return new_counts, labile_modifications


_HEX = FrozenMonosaccharideResidue.from_iupac_lite("Hex")
_HEXNAC = FrozenMonosaccharideResidue.from_iupac_lite("HexNAc")
_FUC = FrozenMonosaccharideResidue.from_iupac_lite("Fuc")
_DHEX = FrozenMonosaccharideResidue.from_iupac_lite("dHex")
_XYL = FrozenMonosaccharideResidue.from_iupac_lite("Xyl")
_AHEX = FrozenMonosaccharideResidue.from_iupac_lite("aHex")

_NEUAC = FrozenMonosaccharideResidue.from_iupac_lite("NeuAc")
_NEUGC = FrozenMonosaccharideResidue.from_iupac_lite("NeuGc")
_NEU = FrozenMonosaccharideResidue.from_iupac_lite("Neu")
_KDN = FrozenMonosaccharideResidue.from_iupac_lite("Kdn")


class _GlycanFragmentingStrategyBase:

    def __init__(self, *args, **kwargs):
        pass

    def _guess_max_glycan_cleavages(self) -> int:
        glycans = self.peptide.glycosylation_manager.items()
        counter = 0
        for pos, glycan in glycans:
            rule = glycan.rule
            if not rule.is_composition and not rule.is_core:
                new_count = rule.glycan.count_branches()
                if new_count > counter:
                    counter = new_count
        return counter


class CADFragmentationStrategy(FragmentationStrategyBase, _GlycanFragmentingStrategyBase):
    """
    Generate glycopeptide fragments derived from glycosidic
    bond cleavages as in lower energy CAD fragmentation
    """
    fold_glycan_compositions: bool

    def __init__(self, peptide, fold_glycan_compositions: bool=True, **kwargs):
        super(CADFragmentationStrategy, self).__init__(peptide, **kwargs)
        self.fold_glycan_compositions = fold_glycan_compositions
        self._generator = self._build_fragments()

    def _is_composition_extended(self, composition: HashableGlycanComposition):
        return composition['HexNAc'] <= 2 and composition["Hex"] <= 3

    def _determine_is_extended_from_glycan_compositions(self, compositions: List[HashableGlycanComposition]):
        return any(self._is_composition_extended(c) for c in compositions)

    def _build_fragments(self):
        glycans = self.peptide.glycosylation_manager.items()
        n = len(glycans)
        if self.compute_compositions:
            base_composition: Composition = self.peptide.peptide_composition()
        else:
            base_composition = None
        peptide_mass: float = self.peptide.peptide_backbone_mass
        for i in range(1, n + 1):
            for dissociated in combinations(glycans, i):
                remaining = dict(glycans)
                for position, glycan in dissociated:
                    remaining.pop(position)
                if self.compute_compositions:
                    reference_composition = base_composition.copy()
                else:
                    reference_composition = None

                glycan_Y_ions = defaultdict(lambda: [None])
                for position, glycan in dissociated:
                    rule: Glycosylation = glycan.rule
                    for f in rule.get_fragments(
                        kind="Y",
                        mask_residues=(_NEUAC, _NEUGC, _NEU, _KDN),
                        fold_glycan_compositions=self.fold_glycan_compositions):
                        glycan_Y_ions[position].append(f)

                key_order = list(glycan_Y_ions.keys())
                is_extended = None
                for y_ion_set in product(*glycan_Y_ions.values()):
                    if self.fold_glycan_compositions:
                        y_ion_set = [gc for gc in y_ion_set if gc is not None]
                        if y_ion_set:
                            glycan_composition: HashableGlycanComposition = sum(y_ion_set, HashableGlycanComposition())
                            glycan_composition.set_composition_offset(Composition())
                            name = str(glycan_composition)
                        else:
                            name = None
                            glycan_composition = None
                        is_extended = self._determine_is_extended_from_glycan_compositions(y_ion_set)
                    else:
                        y_ion_set = [f for f in y_ion_set if f is not None]
                        if y_ion_set:
                            name = ",".join("%d:%s" % (k, f.name) for k, f in zip(key_order, y_ion_set))
                        else:
                            name = None
                        glycan_composition = None
                    if self.compute_compositions:
                        fragment_composition = reference_composition.copy()
                    else:
                        fragment_composition = None
                    is_glycosylated = False
                    mass_shift = 0.0
                    if self.fold_glycan_compositions:
                        if glycan_composition is not None:
                            mass_shift = glycan_composition.mass()
                            if self.compute_compositions:
                                fragment_composition += glycan_composition.total_composition()
                        is_glycosylated = bool(y_ion_set)
                    else:
                        for f in y_ion_set:
                            mass_shift += f.mass
                            if self.compute_compositions:
                                fragment_composition += f.composition
                            is_glycosylated = True
                    full_name = "peptide"
                    if name:
                        full_name = "%s+%s" % (full_name, name)
                    if self.fold_glycan_compositions:
                        f = StubFragment(
                            full_name,
                            peptide_mass + mass_shift,
                            IonSeries.stub_glycopeptide,
                            fragment_composition,
                            glycosylation=glycan_composition,
                            is_extended=is_extended,
                        )
                    else:
                        f = SimpleFragment(
                            name=full_name,
                            mass=peptide_mass + mass_shift,
                            composition=fragment_composition,
                            is_glycosylated=is_glycosylated,
                            kind=IonSeries.stub_glycopeptide)
                    yield f

    def __next__(self):
        return next(self._generator)

    def next(self):
        return self.__next__()

    def __iter__(self):
        return self


class _MonosaccharideDefinitionCacher(object):
    def __init__(self, *args, **kwargs):
        self._hexose = None
        self._hexnac = None
        self._neuac = None
        self._neugc = None
        self._fucose = None
        self._dhex = None
        self._xylose = None
        self._hexa = None
        super(_MonosaccharideDefinitionCacher, self).__init__(*args, **kwargs)

    def _prepare_monosaccharide(self, name):
        return FrozenMonosaccharideResidue.from_iupac_lite(name)

    @property
    def hexose(self):
        if self._hexose is None:
            self._hexose = self._prepare_monosaccharide("Hex")
        return self._hexose

    @property
    def hexnac(self):
        if self._hexnac is None:
            self._hexnac = self._prepare_monosaccharide("HexNAc")
        return self._hexnac

    @property
    def fucose(self):
        if self._fucose is None:
            self._fucose = self._prepare_monosaccharide("Fuc")
        return self._fucose

    @property
    def dhex(self):
        if self._dhex is None:
            self._dhex = self._prepare_monosaccharide("dHex")
        return self._dhex

    @property
    def neuac(self):
        if self._neuac is None:
            self._neuac = self._prepare_monosaccharide("Neu5Ac")
        return self._neuac

    @property
    def neugc(self):
        if self._neugc is None:
            self._neugc = self._prepare_monosaccharide("Neu5Gc")
        return self._neugc

    @property
    def xylose(self):
        if self._xylose is None:
            self._xylose = self._prepare_monosaccharide("Xyl")
        return self._xylose

    @property
    def hexa(self):
        if self._hexa is None:
            self._hexa = self._prepare_monosaccharide("HexA")
        return self._hexa


_Hex2NAc = FrozenMonosaccharideResidue.from_iupac_lite("Hex2NAc")
_Glc2NAc = FrozenMonosaccharideResidue.from_iupac_lite("Glc2NAc")
_Gal2NAc = FrozenMonosaccharideResidue.from_iupac_lite("Gal2NAc")


class GlycanCompositionFragmentStrategyBase(FragmentationStrategyBase):
    def __init__(self, peptide, use_query=False, *args, **kwargs):
        super(GlycanCompositionFragmentStrategyBase, self).__init__(peptide, *args, **kwargs)
        # to be initialized closer to when the glycan composition will
        # be used. Determine whether to use the fast-path __getitem__ or
        # go through the slower but more general GlycanComposition.query
        # method to look up monosaccharides.
        self._use_query = use_query
        self._generator = None

    def glycan_composition(self):
        return self.peptide.glycan_composition

    def _guess_query_mode(self, glycan_composition):
        # these guesses will work for N-glycans and common types of mucin-type O-glycans
        # and GAG linkers
        flag = glycan_composition._getitem_fast(_Hex2NAc) +\
            glycan_composition._getitem_fast(_Glc2NAc) +\
            glycan_composition._getitem_fast(_Gal2NAc)
        return flag or self._use_query

    def count_glycosylation_type(self, glycotype):
        return self.peptide.glycosylation_manager.count_glycosylation_type(glycotype)



class GlycanCompositionFragment(object):
    __slots__ = ("mass", "composition", "key", "is_extended", "_hash_key")

    def __init__(self, mass, composition, key, is_extended=False):
        self.mass = mass
        self.composition = composition
        self.key = key
        self.is_extended = is_extended
        self._hash_key = -1

    def __getitem__(self, key):
        if key == "mass":
            return self.mass
        elif key == "composition":
            return self.composition
        elif key == "key":
            return self.key
        elif key == 'is_extended':
            return self.is_extended
        else:
            raise KeyError(key)

    def __hash__(self):
        if self._hash_key == -1:
            self._hash_key = hash(self._get_glycan_composition())
        return self._hash_key

    def _get_glycan_composition(self):
        return _prepare_glycan_composition_from_mapping(self.key)

    def __eq__(self, other):
        return self.key == other.key and self.is_extended == other.is_extended


class _CompositionTree(object):
    def __init__(self, root=None):
        if root is None:
            root = _NameTree()
        self.root = root

    def __getitem__(self, key):
        return self.root[key]

    def build(self, pair_sequence):
        node = self.root.traverse(pair_sequence)
        if node.name is None:
            # modify the hashable glycan composition before calculating
            # its hash value
            node.name = HashableGlycanComposition()
            for k, v in pair_sequence:
                node.name._setitem_fast(k, v)
            # Force the population of the _mass and _str caches
            node.name.mass()
            str(node.name)
        return node.name


_composition_tree_root = _CompositionTree()
_composition_name_cache = dict()


def _prepare_glycan_composition_from_mapping(mapping):
    pair_sequence = mapping.items()
    return _composition_tree_root.build(pair_sequence)


class StubGlycopeptideStrategy(GlycanCompositionFragmentStrategyBase, _MonosaccharideDefinitionCacher):

    """A fragmentation strategy that generates intact peptide + glycan Y fragments from
    glycopeptides where the glycan structure is not fully known. When the structure is
    known, :class:`CADFragmentationStrategy` should be used instead

    Attributes
    ----------
    extended : bool
        Whether to fragment beyond the conserved core
    extended_fucosylation : bool
        Whether to consider multiple ``Fuc`` residues per fragment rather than one per core
    """

    def __init__(self, peptide, extended=True, use_query=False, extended_fucosylation=False, **kwargs):
        self.extended = extended
        self.extended_fucosylation = extended_fucosylation
        super(StubGlycopeptideStrategy, self).__init__(peptide, use_query)

    def __next__(self):
        if self._generator is None:
            self._generator = self.stub_fragments()
        return next(self._generator)

    def fucosylate_increment(self, shift):
        fucosylated = shift.copy()
        fucosylated['key'] = fucosylated['key'].copy()
        fucosylated['mass'] += self.fucose.mass()
        if self.compute_compositions:
            fucosylated['composition'] = fucosylated[
                'composition'] + self.fucose.total_composition()
        fucosylated['key'][_FUC] = 1
        return fucosylated

    def xylosylate_increment(self, shift):
        xylosylated = shift.copy()
        xylosylated['key'] = xylosylated['key'].copy()
        xylosylated['mass'] += self.xylose.mass()
        if self.compute_compositions:
            xylosylated['composition'] = xylosylated[
                'composition'] + self.xylose.total_composition()
        xylosylated['key'][_XYL] = 1
        return xylosylated

    def fucosylate_extended(self, shift, fucose_count):
        result = [None for i in range(fucose_count)]
        for i in range(1, fucose_count + 1):
            fucosylated = shift.copy()
            fucosylated['key'] = fucosylated['key'].copy()
            fucosylated['mass'] += self.fucose.mass()
            if self.compute_compositions:
                fucosylated['composition'] = fucosylated[
                    'composition'] + self.fucose.total_composition() * i
            fucosylated['key'][_FUC] = i
            result[i - 1] = fucosylated
        return result

    def modified_increment(self, modified, shift):
        modified = modified.copy()
        modified['key'] = modified['key'].copy()
        modified['mass'] += shift.mass()
        if self.compute_compositions:
            modified['composition'] = modified['composition'] + \
                shift.total_composition()
        modified['key'][shift] = 1
        return modified

    def _validate_glycan_composition(self, aggregate_glycosylation, glycan):
        invalid = False
        if self._use_query:
            for key, value in aggregate_glycosylation.items():
                if glycan.query(key) < value:
                    invalid = True
                    break
        else:
            for key, value in aggregate_glycosylation.items():
                if glycan[key] < value:
                    invalid = True
                    break
        return invalid

    def n_glycan_stub_fragments(self):
        glycan = self.glycan_composition()
        self._use_query = self._guess_query_mode(glycan)
        core_count = self.count_glycosylation_type(GlycosylationType.n_linked)
        per_site_shifts = []
        base_composition = self.peptide_composition()
        base_mass = base_composition.mass
        seen = set()
        for i in range(core_count):
            core_shifts = self.n_glycan_composition_fragments(glycan, core_count, i)
            per_site_shifts.append(core_shifts)
        for positions in product(*per_site_shifts):
            aggregate_glycosylation = _AccumulatorBag()
            mass = base_mass
            composition = base_composition.clone()
            is_extended = False
            n_positions = len(positions)
            for site in positions:
                mass += site['mass']
                if n_positions > 1:
                    aggregate_glycosylation += (site['key'])
                else:
                    aggregate_glycosylation = site['key']
                composition += site['composition']
                is_extended |= site['is_extended']
            is_glycosylated = (mass != base_mass)
            if len(positions) > 1:
                invalid = self._validate_glycan_composition(aggregate_glycosylation, glycan)
                if invalid:
                    continue
            glycosylation = _prepare_glycan_composition_from_mapping(
                aggregate_glycosylation)
            name_key = str(glycosylation)
            try:
                name = _composition_name_cache[name_key]
            except KeyError:
                _composition_name_cache[name_key] = name = StubFragment.build_name_from_composition(
                    aggregate_glycosylation)
            if name in seen:
                continue
            seen.add(name)
            yield StubFragment(
                name=name,
                mass=mass,
                composition=composition,
                is_glycosylated=is_glycosylated,
                kind=IonSeries.stub_glycopeptide,
                glycosylation=glycosylation,
                is_extended=is_extended)

    def n_glycan_composition_fragments(self, glycan, core_count=1, iteration_count=0):
        """Generate theoretical N-glycan Y fragment compositions containing the core motif
        plus a portion of the extended branches if :attr:`extended` is used. Unless :attr:`extend

        Parameters
        ----------
        glycan : :class:`~.GlycanComposition`
            Description
        core_count : int, optional
            The total number of glycans attached to the peptide
        iteration_count : int, optional
            The core index. Unless :attr:`extended_fucosylation` is true, this will
            limit the number of Fucose residues to one per core at most, and if
            the iteration count exceeds the number of cores no Fucose will be added

        Returns
        -------
        list of dict
            The compositions corresponding to the theoretical fragments
        """
        hexose = self.hexose
        hexnac = self.hexnac

        if self._use_query:
            fucose_count = glycan.query('Fuc') + glycan.query('dHex')
            xylose_count = glycan.query('Xyl')
            hexnac_in_aggregate = glycan.query('HexNAc')
            hexose_in_aggregate = glycan.query('Hex')
        else:
            fucose_count = glycan[_FUC] + glycan[_DHEX]
            xylose_count = glycan[_XYL]
            hexnac_in_aggregate = glycan[_HEXNAC]
            hexose_in_aggregate = glycan[_HEX]

        core_shifts = []
        base_hexnac = min(hexnac_in_aggregate + 1, 3)
        for hexnac_count in range(base_hexnac):
            if hexnac_count == 0:
                shift = {
                    "mass": 0,
                    "composition": Composition(),
                    "key": {},
                    'is_extended': False,
                }
                core_shifts.append(shift)
            elif hexnac_count == 1:
                shift = {
                    "mass": (hexnac_count * hexnac.mass()),
                    "composition": hexnac_count * hexnac.total_composition(),
                    "key": {_HEXNAC: hexnac_count},
                    'is_extended': False,
                }
                core_shifts.append(shift)
                if iteration_count < fucose_count:
                    fucosylated = self.fucosylate_increment(shift)
                    core_shifts.append(fucosylated)
            elif hexnac_count == 2:
                shift = {
                    "mass": (hexnac_count * hexnac.mass()),
                    "composition": hexnac_count * hexnac.total_composition(),
                    "key": {_HEXNAC: hexnac_count},
                    'is_extended': False,
                }
                core_shifts.append(shift)

                if not self.extended_fucosylation:
                    if iteration_count < fucose_count:
                        fucosylated = self.fucosylate_increment(shift)
                        core_shifts.append(fucosylated)
                        if iteration_count < xylose_count:
                            xylosylated = self.xylosylate_increment(fucosylated)
                            core_shifts.append(xylosylated)
                elif fucose_count > 0:
                    core_shifts.extend(self.fucosylate_extended(shift, fucose_count))
                    if iteration_count < xylose_count:
                        xylosylated = self.xylosylate_increment(fucosylated)
                        core_shifts.append(xylosylated)

                if iteration_count < xylose_count:
                    xylosylated = self.xylosylate_increment(shift)
                    core_shifts.append(xylosylated)

                for hexose_count in range(1, min(hexose_in_aggregate + 1, 4)):
                    shift = {
                        "mass": (hexnac_count * hexnac.mass()) + (hexose_count * hexose.mass()),
                        "composition": (hexnac_count * hexnac.total_composition()) + (
                            hexose_count * hexose.total_composition()),
                        "key": {_HEXNAC: hexnac_count, _HEX: hexose_count},
                        'is_extended': False,
                    }
                    core_shifts.append(shift)

                    if not self.extended_fucosylation:
                        if iteration_count < fucose_count:
                            fucosylated = self.fucosylate_increment(shift)
                            core_shifts.append(fucosylated)
                            if iteration_count < xylose_count:
                                xylosylated = self.xylosylate_increment(fucosylated)
                                core_shifts.append(xylosylated)
                    elif fucose_count > 0:
                        core_shifts.extend(self.fucosylate_extended(shift, fucose_count))
                        if iteration_count < xylose_count:
                            xylosylated = self.xylosylate_increment(fucosylated)
                            core_shifts.append(xylosylated)

                    if iteration_count < xylose_count:
                        xylosylated = self.xylosylate_increment(shift)
                        core_shifts.append(xylosylated)

                    # After the core motif has been exhausted, speculatively add
                    # on the remaining core monosaccharides sequentially until
                    # exhausted.
                    if hexose_count == 3 and hexnac_in_aggregate >= 2 * core_count and self.extended:
                        for extra_hexnac_count in range(0, hexnac_in_aggregate - hexnac_count + 1):
                            if extra_hexnac_count + hexnac_count > hexnac_in_aggregate:
                                continue
                            if extra_hexnac_count > 0:
                                shift = {
                                    "mass": (
                                        (hexnac_count + extra_hexnac_count) * hexnac.mass()) + (
                                        hexose_count * hexose.mass()),
                                    "composition": (
                                        (hexnac_count + extra_hexnac_count) * hexnac.total_composition()) + (
                                        hexose_count * hexose.total_composition()),
                                    "key": {_HEXNAC: hexnac_count + extra_hexnac_count, _HEX: hexose_count},
                                    'is_extended': True
                                }
                                core_shifts.append(shift)

                                if not self.extended_fucosylation:
                                    if iteration_count < fucose_count:
                                        fucosylated = self.fucosylate_increment(shift)
                                        core_shifts.append(fucosylated)
                                        if iteration_count < xylose_count:
                                            xylosylated = self.xylosylate_increment(fucosylated)
                                            core_shifts.append(xylosylated)
                                elif fucose_count > 0:
                                    core_shifts.extend(self.fucosylate_extended(shift, fucose_count))

                                if iteration_count < xylose_count:
                                    xylosylated = self.xylosylate_increment(shift)
                                    core_shifts.append(xylosylated)

                            for extra_hexose_count in range(1, hexose_in_aggregate - hexose_count + 1):
                                if extra_hexose_count + hexose_count > hexose_in_aggregate:
                                    continue
                                shift = {
                                    "mass": (
                                        (hexnac_count + extra_hexnac_count) * hexnac.mass()) + (
                                        (hexose_count + extra_hexose_count) * hexose.mass()),
                                    "composition": (
                                        (hexnac_count + extra_hexnac_count) * hexnac.total_composition()) + (
                                        (hexose_count + extra_hexose_count) * hexose.total_composition()),
                                    "key": {_HEXNAC: hexnac_count + extra_hexnac_count, _HEX: (
                                        hexose_count + extra_hexose_count)},
                                    'is_extended': True
                                }
                                core_shifts.append(shift)

                                if not self.extended_fucosylation:
                                    if iteration_count < fucose_count:
                                        fucosylated = self.fucosylate_increment(shift)
                                        core_shifts.append(fucosylated)
                                        if iteration_count < xylose_count:
                                            xylosylated = self.xylosylate_increment(fucosylated)
                                            core_shifts.append(xylosylated)
                                elif fucose_count > 0:
                                    core_shifts.extend(self.fucosylate_extended(shift, fucose_count))
                                    if iteration_count < xylose_count:
                                        xylosylated = self.xylosylate_increment(fucosylated)
                                        core_shifts.append(xylosylated)

                                if iteration_count < xylose_count:
                                    xylosylated = self.xylosylate_increment(shift)
                                    core_shifts.append(xylosylated)
        return core_shifts

    def o_glycan_stub_fragments(self):
        glycan = self.glycan_composition()
        self._use_query = self._guess_query_mode(glycan)
        core_count = self.count_glycosylation_type(GlycosylationType.o_linked)
        per_site_shifts = []

        base_composition = self.peptide_composition()
        base_mass = base_composition.mass

        for i in range(core_count):
            core_shifts = self.o_glycan_composition_fragments(glycan, core_count, i)
            per_site_shifts.append(core_shifts)
        seen = set()
        for positions in product(*per_site_shifts):
            aggregate_glycosylation = _AccumulatorBag()
            mass = base_mass
            composition = base_composition.clone()
            for site in positions:
                mass += site['mass']
                aggregate_glycosylation += (site['key'])
                composition += site['composition']
            is_glycosylated = (composition != base_composition)
            invalid = False
            if self._use_query:
                for key, value in aggregate_glycosylation.items():
                    if glycan.query(key) < value:
                        invalid = True
                        break
            else:
                for key, value in aggregate_glycosylation.items():
                    if glycan[key] < value:
                        invalid = True
                        break
            if invalid:
                continue
            name = StubFragment.build_name_from_composition(aggregate_glycosylation)
            if name in seen:
                continue
            seen.add(name)
            glycosylation = _prepare_glycan_composition_from_mapping(
                aggregate_glycosylation)
            yield StubFragment(
                name=name,
                mass=mass,
                composition=composition,
                is_glycosylated=is_glycosylated,
                kind=IonSeries.stub_glycopeptide,
                glycosylation=glycosylation)

    def o_glycan_composition_fragments(self, glycan, core_count=1, iteration_count=0):
        hexose = self.hexose
        hexnac = self.hexnac

        if self._use_query:
            fucose_count = glycan.query('Fuc') + glycan.query('dHex')
            hexnac_in_aggregate = glycan.query('HexNAc')
            hexose_in_aggregate = glycan.query('Hex')
        else:
            fucose_count = glycan[_FUC] + glycan[_DHEX]
            hexnac_in_aggregate = glycan[_HEXNAC]
            hexose_in_aggregate = glycan[_HEX]
        core_shifts = []
        for hexnac_count in range(3):
            if hexnac_in_aggregate < hexnac_count:
                continue
            if hexnac_count == 0:
                shift = {
                    "mass": 0,
                    "composition": Composition(),
                    "key": {},
                    "is_extended": False
                }
                core_shifts.append(shift)
            elif hexnac_count >= 1:
                shift = {
                    "mass": (hexnac_count * hexnac.mass()),
                    "composition": hexnac_count * hexnac.total_composition(),
                    "key": {_HEXNAC: hexnac_count},
                    "is_extended": False
                }
                core_shifts.append(shift)
                if iteration_count < fucose_count:
                    fucosylated = self.fucosylate_increment(shift)
                    core_shifts.append(fucosylated)
                for hexose_count in range(0, 2):
                    if hexose_in_aggregate < hexose_count:
                        continue
                    if hexose_count > 0:
                        shift = {
                            "mass": (
                                (hexnac_count) * hexnac.mass()) + (
                                (hexose_count) * hexose.mass()),
                            "composition": (
                                (hexnac_count) * hexnac.total_composition()) + (
                                (hexose_count) * hexose.total_composition()),
                            "key": {_HEXNAC: hexnac_count, _HEX: (
                                hexose_count)},
                            "is_extended": False
                        }
                        core_shifts.append(shift)
                        if iteration_count < fucose_count:
                            fucosylated = self.fucosylate_increment(shift)
                            core_shifts.append(fucosylated)
                    # After the core motif has been exhausted, speculatively add
                    # on the remaining core monosaccharides sequentially until
                    # exhausted.
                    if self.extended and hexnac_in_aggregate - hexnac_count >= 0:
                        for extra_hexnac_count in range(hexnac_in_aggregate - hexnac_count + 1):
                            if extra_hexnac_count:
                                shift = {
                                    "mass": (
                                        (hexnac_count + extra_hexnac_count) * hexnac.mass()) + (
                                        (hexose_count) * hexose.mass()),
                                    "composition": (
                                        (hexnac_count + extra_hexnac_count) * hexnac.total_composition()) + (
                                        (hexose_count) * hexose.total_composition()),
                                    "key": {_HEXNAC: hexnac_count + extra_hexnac_count, _HEX: (
                                        hexose_count)},
                                    'is_extended': True,
                                }
                                core_shifts.append(shift)
                                if iteration_count < fucose_count:
                                    fucosylated = self.fucosylate_increment(shift)
                                    core_shifts.append(fucosylated)
                            if hexose_in_aggregate > hexose_count and hexose_count > 0:
                                for extra_hexose_count in range(hexose_in_aggregate - hexose_count):
                                    if extra_hexose_count > 0 and extra_hexose_count + hexose_count > 0:
                                        shift = {
                                            "mass": (
                                                (hexnac_count + extra_hexnac_count) * hexnac.mass()) + (
                                                (hexose_count + extra_hexose_count) * hexose.mass()),
                                            "composition": (
                                                (hexnac_count + extra_hexnac_count) * hexnac.total_composition()) + (
                                                (hexose_count + extra_hexose_count) * hexose.total_composition()),
                                            "key": {_HEXNAC: hexnac_count + extra_hexnac_count, _HEX: (
                                                hexose_count + extra_hexose_count)},
                                            "is_extended": True
                                        }
                                        core_shifts.append(shift)
                                        if iteration_count < fucose_count:
                                            fucosylated = self.fucosylate_increment(
                                                shift)
                                            core_shifts.append(fucosylated)

        return core_shifts

    def gag_linker_stub_fragments(self):
        glycan = self.glycan_composition()
        self._use_query = self._guess_query_mode(glycan)
        core_count = self.count_glycosylation_type(GlycosylationType.glycosaminoglycan)

        per_site_shifts = []

        base_composition = self.peptide_composition()
        base_mass = base_composition.mass
        for i in range(core_count):
            core_shifts = self.gag_linker_composition_fragments(glycan, core_count, i)
            per_site_shifts.append(core_shifts)
        seen = set()
        for positions in product(*per_site_shifts):
            aggregate_glycosylation = _AccumulatorBag()
            mass = base_mass
            composition = base_composition.clone()
            for site in positions:
                mass += site['mass']
                aggregate_glycosylation += (site['key'])
                composition += site['composition']
            is_glycosylated = (composition != base_composition)
            invalid = False
            if self._use_query:
                for key, value in aggregate_glycosylation.items():
                    if glycan.query(key) < value:
                        invalid = True
                        break
            else:
                for key, value in aggregate_glycosylation.items():
                    if glycan[key] < value:
                        invalid = True
                        break
            if invalid:
                continue
            name = StubFragment.build_name_from_composition(aggregate_glycosylation)
            if name in seen:
                continue
            seen.add(name)
            glycosylation = _prepare_glycan_composition_from_mapping(
                aggregate_glycosylation)
            yield StubFragment(
                name=name,
                mass=mass,
                composition=composition,
                is_glycosylated=is_glycosylated,
                kind=IonSeries.stub_glycopeptide,
                glycosylation=glycosylation)

    def gag_linker_composition_fragments(self, glycan, core_count=1, iteration_count=0):
        hexose = self.hexose
        xyl = self.xylose
        hexa = self.hexa

        xyl_in_aggregate = glycan[xyl]
        if xyl_in_aggregate == 0:
            xyl_in_aggregate = glycan.query("Pen", exact=False)
        core_shifts = []
        for xyl_count in range(max(0, xyl_in_aggregate) + 1):
            if xyl_count == 0:
                shift = {
                    "mass": 0,
                    "composition": Composition(),
                    "key": {},
                    "is_extended": False
                }
                core_shifts.append(shift)
            else:
                shift = {
                    "mass": xyl.mass() * xyl_count,
                    "composition": xyl.total_composition() * xyl_count,
                    "key": {
                        _XYL: xyl_count
                    },
                    "is_extended": False
                }
                core_shifts.append(shift)
            if xyl_count > 0:
                # TODO: Handle modified Hexose residues here too.
                for hexose_count in range(1, 3):
                    shift = {
                        "mass": ((xyl.mass() * xyl_count) + (
                            hexose.mass() * hexose_count)),
                        "composition": (
                            (xyl.total_composition() * xyl_count) + (
                                hexose.total_composition() * hexose_count)),
                        "key": {
                            _XYL: xyl_count,
                            _HEX: hexose_count
                        },
                        "is_extended": False
                    }
                    core_shifts.append(shift)
                    if hexose_count == 2:
                        shift = {
                            "mass": ((xyl.mass() * xyl_count) + (
                                hexose.mass() * hexose_count) + hexa.mass()),
                            "composition": (
                                (xyl.total_composition() * xyl_count) + (
                                    hexose.total_composition() * hexose_count) + hexa.total_composition()),
                            "key": {
                                _XYL: xyl_count,
                                _HEX: hexose_count,
                                _AHEX: 1
                            },
                            "is_extended": False
                        }
                        core_shifts.append(shift)
        return core_shifts

    def stub_fragments(self):
        n_glycan = self.count_glycosylation_type(GlycosylationType.n_linked) > 0
        o_glycan = self.count_glycosylation_type(GlycosylationType.o_linked) > 0
        gag_linker = self.count_glycosylation_type(GlycosylationType.glycosaminoglycan) > 0
        if (n_glycan + o_glycan + gag_linker) > 2:
            raise ValueError(
                "Does not support mixed-type glycan fragmentation (yet)")
        if n_glycan:
            return self.n_glycan_stub_fragments()
        elif o_glycan:
            return self.o_glycan_stub_fragments()
        elif gag_linker:
            return self.gag_linker_stub_fragments()
        else:
            if len(self.peptide.glycosylation_manager) > 0:
                raise ValueError("Unknown Glycan Class Detected")
        return (a for a in [])


try:
    _StubGlycopeptideStrategy = StubGlycopeptideStrategy
    _GlycanCompositionFragment = GlycanCompositionFragment

    from glycopeptidepy._c.structure.fragmentation_strategy.glycan import (
        GlycanCompositionFragmentStrategyBase, StubGlycopeptideStrategy, GlycanCompositionFragment)
    _has_c = True
except ImportError:
    _has_c = False


_StubGlycopeptideStrategy = StubGlycopeptideStrategy


class LabileAwareStubGlycopeptideStrategy(_StubGlycopeptideStrategy):
    def __init__(self, peptide, extended=True, use_query=False, extended_fucosylation=False, detatch_substituents=False, **kwargs):
        self.detatch_substituents = detatch_substituents
        self.labile_modifications = None
        super(LabileAwareStubGlycopeptideStrategy,
              self).__init__(peptide, use_query=use_query, extended=extended, extended_fucosylation=extended_fucosylation, **kwargs)

    def glycan_composition(self):
        gc = super(LabileAwareStubGlycopeptideStrategy, self).glycan_composition()
        if self.detatch_substituents:
            gc, self.labile_modifications = separate_labile_modifications(gc)
            return gc
        else:
            return gc


labile_monosaccharides_not_to_duplicate = {
    FrozenMonosaccharideResidue.from_iupac_lite("NeuAc"),
    FrozenMonosaccharideResidue.from_iupac_lite("NeuGc"),
    FrozenMonosaccharideResidue.from_iupac_lite("Neu"),
    FrozenMonosaccharideResidue.from_iupac_lite("Fuc"),
    FrozenMonosaccharideResidue.from_iupac_lite("dHex"),
}

WATER = Composition("H2O")
SIDE_CHAIN_PLUS_CARBON = Composition("CH2O")
TWO_SIDE_CHAIN_PLUS_CARBON = SIDE_CHAIN_PLUS_CARBON * 2
TWO_WATER = WATER * 2


WATER_LOSS = NeutralLoss("-" + formula(WATER), -WATER)
SIDE_CHAIN_LOSS = NeutralLoss(
    "-" + formula(SIDE_CHAIN_PLUS_CARBON), -SIDE_CHAIN_PLUS_CARBON
)

TWO_SIDE_CHAIN_LOSS = NeutralLoss(
    "-" + formula(2 * SIDE_CHAIN_PLUS_CARBON), -2 * SIDE_CHAIN_PLUS_CARBON
)

TWO_WATER_LOSS = NeutralLoss("-" + formula(TWO_WATER), -TWO_WATER)
TWO_WATER_PLUS_SIDE_CHAIN_LOSS = NeutralLoss(
    "-" + formula((TWO_WATER + SIDE_CHAIN_PLUS_CARBON)),
    -(TWO_WATER + SIDE_CHAIN_PLUS_CARBON),
)

TWO_WATER_PLUS_TWO_SIDE_CHAIN_LOSS = NeutralLoss(
    "-" + formula((TWO_WATER + TWO_SIDE_CHAIN_PLUS_CARBON)),
    -(TWO_WATER + TWO_SIDE_CHAIN_PLUS_CARBON),
)

class OxoniumIonStrategy(GlycanCompositionFragmentStrategyBase, _MonosaccharideDefinitionCacher):

    def __init__(self, peptide, use_query=False, oxonium=True, all_series=False, allow_ambiguous=False,
                 include_large_glycan_fragments=True, maximum_fragment_size=5, *args, **kwargs):
        self.oxonium = oxonium
        self.all_series = all_series
        self.allow_ambiguous = allow_ambiguous
        self.maximum_fragment_size = maximum_fragment_size
        self.include_large_glycan_fragments = include_large_glycan_fragments
        super(OxoniumIonStrategy, self).__init__(peptide, use_query=use_query, *args, **kwargs)

    def __next__(self):
        if self._generator is None:
            self._generator = self._build_generator()
        return next(self._generator)

    def _oxonium_fragments_get_monosaccharide_list(self, glycan):
        monosaccharides = dict(glycan)
        for mono, count in list(monosaccharides.items()):
            dissociated, lost_mods = remove_labile_modifications(mono)
            if dissociated != mono:
                monosaccharides[dissociated] = count
                for mod in lost_mods:
                    pass
        return monosaccharides

    def _glycan_structural_dissociation(self, max_cleavages=2):
        # return CADFragmentationStrategy(self.peptide, max_cleavages)
        raise NotImplementedError()

    def _oxonium_ions(self):
        glycan = (self.glycan_composition()).clone()

        monosaccharides = self._oxonium_fragments_get_monosaccharide_list(glycan)
        for k in monosaccharides:
            key = str(k)
            mass = k.mass()
            composition = k.total_composition()
            base = OxoniumIon(
                name=key, mass=mass,
                composition=composition,
                kind=oxonium_ion_series,
                monosaccharides=[k]
            )

            yield base

            loss: OxoniumIon = base.clone()
            loss.set_chemical_shift(WATER_LOSS)
            yield loss

            loss: OxoniumIon = base.clone()
            loss.set_chemical_shift(TWO_WATER_LOSS)
            yield loss

            loss: OxoniumIon = base.clone()
            loss.set_chemical_shift(TWO_SIDE_CHAIN_LOSS)
            yield loss

            loss: OxoniumIon = base.clone()
            loss.set_chemical_shift(TWO_WATER_PLUS_SIDE_CHAIN_LOSS)
            yield loss

            loss: OxoniumIon = base.clone()
            loss.set_chemical_shift(TWO_WATER_PLUS_TWO_SIDE_CHAIN_LOSS)
            yield loss

        for i in range(2, self.maximum_fragment_size + 1):
            for kk in combinations_with_replacement(sorted(monosaccharides, key=str), i):
                invalid = False
                for k, v in Counter(kk).items():
                    if monosaccharides[k] < v:
                        invalid = True
                        break
                    elif k in labile_monosaccharides_not_to_duplicate and v > 1:
                        invalid = True
                        break
                if invalid:
                    continue
                key = ''.join(map(str, kk))
                mass = sum(k.mass() for k in kk)
                composition = sum((k.total_composition()
                                   for k in kk), Composition())
                base = OxoniumIon(
                    name=key, mass=mass, kind=oxonium_ion_series, composition=composition,
                    monosaccharides=list(kk)
                )
                yield base

                loss = base.clone()
                loss.set_chemical_shift(WATER_LOSS)
                yield loss

                loss = base.clone()
                loss.set_chemical_shift(TWO_WATER_LOSS)
                yield loss

    def _ambiguous_all_series(self):
        water = Composition("H2O")
        _hexnac = self.hexnac
        _hexose = self.hexose
        _neuac = self.neuac

        if self.peptide.glycosylation_manager.count_glycosylation_type(GlycosylationType.n_linked) > 0:
            _offset = Composition()
            total = (self.glycan_composition()).clone()
            total_count = sum(total.values())

            base = FrozenGlycanComposition(Hex=3, HexNAc=2)
            remainder = total - base

            peptide_base_composition = self.total_composition() - total.total_composition()
            stub_composition = peptide_base_composition + base.total_composition() - water
            stub_mass = stub_composition.mass

            # GlycanComposition's clone semantics do not propagate the
            # composition_offset attribute yet. Should it?
            remainder.composition_offset = _offset
            remainder_elemental_composition = remainder.total_composition()
            remainder_mass = remainder.mass()

            for composition in descending_combination_counter(remainder):
                frag_size = sum(composition.values())

                # Don't waste time on compositions that are impossible under
                # common enzymatic pathways in this already questionable
                # stop-gap
                if composition.get(_hexnac, 0) + composition.get(
                        _hexose, 0) < composition.get(_neuac, 0):
                    continue

                composition = FrozenGlycanComposition(composition)
                composition.composition_offset = _offset

                elemental_composition = composition.total_composition()
                composition_mass = elemental_composition.mass

                if frag_size > 2 and self.include_large_glycan_fragments and frag_size < self.maximum_fragment_size:
                    string_form = composition.serialize()
                    yield SimpleFragment(
                        name=string_form, mass=composition_mass,
                        composition=elemental_composition, kind=oxonium_ion_series)
                    yield SimpleFragment(
                        name=string_form + "-H2O", mass=composition_mass - water.mass,
                        composition=elemental_composition - water, kind=oxonium_ion_series)

                if (total_count - frag_size) < (self.maximum_fragment_size + 4):
                    f = StubFragment(
                        name="peptide+" + str(total - composition),
                        mass=stub_mass + remainder_mass - composition_mass,
                        composition=stub_composition + remainder_elemental_composition - elemental_composition,
                        is_glycosylated=True,
                        kind=IonSeries.stub_glycopeptide,
                        glycosylation=composition,
                        is_extended=None)
                    yield f

    def _build_generator(self):
        if self.oxonium:
            for f in self._oxonium_ions():
                yield f
        if self.peptide.glycosylation_manager.is_fully_specified_topologies() and self.all_series:
            for f in self._glycan_structural_dissociation():
                yield f
        elif self.allow_ambiguous and self.all_series:
            for f in self._ambiguous_all_series():
                yield f
        elif self.all_series:
            raise TypeError(
                "Cannot generate B/Y fragments from non-Glycan {}".format(self.glycan))



from typing import Dict, DefaultDict, List, Union


from glycopeptidepy.structure import ProteinSequence, ModificationTable, ModificationRule

from .fasta import PEFFFeature, PEFFFastaHeader
from . import uniprot
from .cv.uniprot_ptm import load as load_uniprot_ptms


class UniprotToPeffConverter(object):

    feature_to_peff_term = {
        uniprot.SignalPeptide.feature_type: ('Processed', 'PEFF:0001021'),
        uniprot.MatureProtein.feature_type: ('Processed', 'PEFF:0001020'),
        uniprot.TransitPeptide.feature_type: ('Processed', 'PEFF:0001022'),
        uniprot.Propeptide.feature_type: ('Processed', 'PEFF:0001034'),
        uniprot.SequenceVariant.feature_type: ('VariantSimple', 'PEFF:0001028'),
        uniprot.GlycosylationSite.feature_type: ('ModRes', ''),
        uniprot.ModifiedResidue.feature_type: ('ModResPsi', '')
    }

    uniprot_ptms: ModificationTable
    psimod_rules: Dict[str, ModificationRule]

    def __init__(self):
        self.uniprot_ptms = load_uniprot_ptms()
        self.psimod_rules = {}
        self.prepare_psimod_conversion_table()

    def prepare_psimod_conversion_table(self):
        psimod_mapping = {}
        for mod in self.uniprot_ptms.rules():

            mod_names = [name for name in mod.names if name.startswith('MOD:')]
            if mod_names:
                psimod_mapping[mod.name] = mod_names[0]
        self.psimod_rules = psimod_mapping

    def handle_Processed(self, feature: uniprot.PeptideBase, accession) -> PEFFFeature:
        peff_feature = PEFFFeature(feature.start + 1, feature.end, accession, feature.feature_type)
        return peff_feature

    def handle_VariantSimple(self, feature, accession) -> PEFFFeature:
        peff_feature = PEFFFeature(feature.position + 1, feature.original, feature.variation)
        return peff_feature

    def handle_ModRes(self, feature: uniprot.ModifiedResidue, accession) -> PEFFFeature:
        if accession:
            peff_feature = PEFFFeature(
                feature.position + 1, str(feature.description), accession)
        else:
            peff_feature = PEFFFeature(feature.position + 1, str(feature.description))
        return peff_feature

    def handle_ModResPsi(self, feature: uniprot.ModifiedResidue, accession) -> PEFFFeature:
        if feature.description in self.psimod_rules:
            accession = self.psimod_rules[feature.description]
            return PEFFFeature(feature.position + 1, accession, feature.description)

    def base_features(self, record: uniprot.UniProtProtein) -> DefaultDict[str, Union[str, List[PEFFFeature]]]:
        features = DefaultDict(list)
        features['Prefix'] = 'sp'
        features['Tag'] = record.accessions[0]
        features['PName'] = record.recommended_name
        features['GName'] = record.gene_name
        return features

    def __call__(self, record: uniprot.UniProtProtein):
        features = self.base_features(record)
        for feature in record.features:
            try:
                key = self.feature_to_peff_term[feature.feature_type]
                handler = getattr(self, 'handle_' + key[0], None)
                if handler is None:
                    continue
                f = handler(feature, key[1])
                if f is None:
                    continue
                features[key[0]].append(f)
            except KeyError:
                continue
        header = PEFFFastaHeader(dict(features))
        return ProteinSequence(header, record.sequence)



cimport cython

from cpython.ref cimport Py_INCREF
from cpython cimport PyObject
from cpython.list cimport PyList_GetItem, PyList_SetItem, PyList_Size, PyList_New, PyList_GET_ITEM, PyList_GET_SIZE
from cpython.dict cimport PyDict_GetItem, PyDict_SetItem
from cpython.int cimport PyInt_AsLong
from cpython.float cimport PyFloat_AsDouble

from glycopeptidepy._c.structure.base cimport (
    PeptideSequenceBase,
    SequencePosition, TerminalGroup, AminoAcidResidueBase,
    ModificationBase)

from glycopeptidepy._c.structure.modification.source cimport ModificationTableBase
from glycopeptidepy._c.structure.modification.modification cimport ModificationInstanceBase

from glycopeptidepy._c.parser import sequence_tokenizer

from glypy.composition.ccomposition cimport CComposition

from glycopeptidepy._c.structure.constants cimport Configuration
from glycopeptidepy._c.structure.glycan cimport GlycosylationManager, GlycanCompositionWithOffsetProxy

from glycopeptidepy.structure.residue import AminoAcidResidue
from glycopeptidepy.structure.modification import Modification, ModificationCategory, SequenceLocation
from glycopeptidepy.structure import constants as _structure_constants

cdef CComposition WATER = CComposition("H2O")

cdef Configuration structure_constants = _structure_constants

cdef object ModificationImpl = Modification
cdef object AminoAcidResidueImpl = AminoAcidResidue

cdef object SequenceLocation_n_term = SequenceLocation.n_term
cdef object SequenceLocation_c_term = SequenceLocation.c_term

cdef object ModificationCategory_glycosylation = ModificationCategory.glycosylation

cdef ModificationTableBase modification_table = ModificationImpl.modification_table


cpdef ModificationBase get_modification_by_name(basestring name):
    cdef:
        ModificationInstanceBase inst

    rule = modification_table.resolve(name)
    inst = <ModificationInstanceBase>ModificationImpl.__new__(ModificationImpl)
    inst._init_from_rule(rule)
    return inst


cdef dict terminal_group_cache = dict()

# Helpers and Caches -----------------

cdef TerminalGroup _make_terminal_group(basestring base_composition_formula, ModificationBase modification=None):
    """A cache for creating :class:`~.TerminalGroup` instances from base formula and
    modification.

    Returns
    -------
    TerminalGroup
    """
    key = (base_composition_formula, modification)
    result = PyDict_GetItem(terminal_group_cache, key)
    if result == NULL:
        inst = TerminalGroup(base_composition_formula, modification)
        PyDict_SetItem(terminal_group_cache, key, inst)
        return inst
    return <TerminalGroup>result


cdef dict amino_acid_cache = dict()


cdef AminoAcidResidueBase _parse_residue(basestring residue_string):
    """A substitute cache for :class:`~.AminoAcidResidue` instantiation.

    Normally, creating an :class:`AminoAcidResidue` instance passes through
    a cache in the class's constructor. This however incurs a lot of Python
    overhead because it must go through the class's ``__call__`` slot *every
    time*, so there are many pointer jumps along the way.

    This function implements a cache in front of that cache because creating
    AminoAcidResidue instances is done so frequently by the :class:`_PeptideSequenceCore`
    initialization code. It just maintains a static mapping between the amino acid symbol
    to the instance of the AminoAcidResidue that represents it, e.g. "N" -> Asparagine.

    Returns
    -------
    AminoAcidResidueBase
    """
    result = PyDict_GetItem(amino_acid_cache, residue_string)
    if result == NULL:
        inst = AminoAcidResidueImpl(residue_string)
        PyDict_SetItem(amino_acid_cache, residue_string, inst)
        return inst
    return <AminoAcidResidueBase>result


cdef dict implicit_terminal_cache = dict()


cdef ModificationBase _get_implicit_terminal(basestring terminal_name):
    """A cache for storing the Modification mapping for implicit N- and C-terminal
    groups

    Returns
    -------
    ModificationBase
    """
    result = PyDict_GetItem(implicit_terminal_cache, terminal_name)
    if result == NULL:
        inst = ModificationImpl(terminal_name)
        PyDict_SetItem(implicit_terminal_cache, terminal_name, inst)
        return inst
    return <ModificationBase>result

# -------------------


cdef CComposition _total_composition(_PeptideSequenceCore sequence):
    cdef:
        CComposition total
        size_t i, j, m, n
        SequencePosition position
        ModificationBase mod

    n = sequence.get_size()
    total = CComposition()
    for i in range(n):
        position = sequence.get(i)
        total += position.amino_acid.composition
        m = SequencePosition.get_modification_count(position)
        for j in range(m):
            mod = SequencePosition.get_modification(position, j)
            if mod.is_tracked_for(ModificationCategory_glycosylation):
                continue
            total += mod.composition

    total += sequence._n_term.get_composition()
    total += sequence._c_term.get_composition()
    gc = sequence.glycan_composition
    if gc:
        total += gc.total_composition()
    return total


cdef class _PeptideSequenceCore(PeptideSequenceBase):
    def __init__(self, sequence=None, parser_function=None, **kwargs):
        if parser_function is None:
            parser_function = sequence_tokenizer
        self._initialize_fields()
        if sequence is not None:
            self._init_from_string(sequence, parser_function)

    cpdef _init_from_parsed_string(self, list seq_list, glycan=None, n_term=None, c_term=None):
        cdef:
            size_t i, j, n, m
            list sequence, item, mods, mod_strs
            double mass
            ModificationBase mod
            AminoAcidResidueBase res
            SequencePosition position
            bint has_glycan
            GlycosylationManager glycosylation_manager

        i = 0
        n = PyList_GET_SIZE(seq_list)
        self.sequence = sequence = PyList_New(n)
        glycosylation_manager = <GlycosylationManager>self._glycosylation_manager
        mass = 0
        for i in range(n):
            item = <list>PyList_GET_ITEM(seq_list, i)
            res = _parse_residue(<basestring>PyList_GET_ITEM(item, 0))
            mass += res.mass
            mod_strs = <list>PyList_GET_ITEM(item, 1)
            mods = None
            if mod_strs is not None:
                mods = []
                m = PyList_GET_SIZE(mod_strs)
                for j in range(m):
                    mod_str = <basestring>PyList_GET_ITEM(mod_strs, j)
                    if mod_str != '':
                        mod = get_modification_by_name(mod_str)
                        if mod.is_tracked_for(ModificationCategory_glycosylation):
                            glycosylation_manager[i] = mod
                        mods.append(mod)
                        mass += mod.mass

            position = SequencePosition._create(res, mods)
            Py_INCREF(position)
            PyList_SetItem(sequence, i, position)

        self._mass = mass
        has_glycan = glycan is not None
        if has_glycan:
            glycosylation_manager.set_aggregate(glycan)

        if isinstance(n_term, basestring):
            if n_term != structure_constants.N_TERM_DEFAULT:
                n_term = get_modification_by_name(n_term)
            else:
                n_term = None
        if isinstance(c_term, basestring):
            if c_term != structure_constants.C_TERM_DEFAULT:
                c_term = get_modification_by_name(c_term)
            else:
                c_term = None
        n_term_group = _make_terminal_group(structure_constants.N_TERM_DEFAULT, n_term)
        c_term_group = _make_terminal_group(structure_constants.C_TERM_DEFAULT, c_term)
        self._init_termini(
            n_term_group,
            c_term_group)

    cpdef _init_from_string(self, sequence, parser_function):
        """Initialize a :class:`PeptideSequence` from a parse-able string.

        Parameters
        ----------
        sequence : :class:`str`
            The string to parse
        parser_function : :class:`Callable`
            The parsing function to use
        """
        if sequence is None or sequence == "":
            pass
        else:
            seq_list, _, glycan, n_term, c_term = <tuple>parser_function(
                sequence)
            self._init_from_parsed_string(seq_list, glycan, n_term, c_term)

    cpdef _init_from_components(self, list seq_list, glycan_composition=None, ModificationBase n_term=None, ModificationBase c_term=None):
        """Initialize a :class:`PeptideSequence`'s state from a sequence of (
        :class:`~.AminoAcidResidue`, [:class:`~.Modification`, ]) pairs, with optional extra information
        defining the glycan composition aggregate and the terminal groups

        Parameters
        ----------
        seq_list : :class:`Sequence`
            The sequence of (:class:`~.AminoAcidResidue`, [:class:`~.Modification`, ])
        glycan_composition : :class:`~.GlycanComposition`, optional
            The aggregate glycan composition associated with the sequence not fully specified
            by the modifications in ``seq_list``.
        n_term : :class:`~.Modification`, optional
            The N-terminal modification, if any
        c_term : :class:`~.Modification`, optional
            The C-terminal modification, if any
        """
        cdef:
            size_t i, j, n, m
            list sequence, mods, mod_strs
            SequencePosition item, position
            double mass
            ModificationBase mod
            AminoAcidResidueBase res
            GlycosylationManager glycosylation_manager

        i = 0
        n = PyList_GET_SIZE(seq_list)
        self.sequence = sequence = PyList_New(n)
        mass = 0
        glycosylation_manager = <GlycosylationManager>self._glycosylation_manager
        for i in range(n):
            item = <SequencePosition>PyList_GET_ITEM(seq_list, i)
            mass += item.amino_acid.mass
            if item.modifications is not None:
                m = SequencePosition.get_modification_count(item)
                mods = PyList_New(m)
                for j in range(m):
                    mod = SequencePosition.get_modification(item, j)
                    Py_INCREF(mod)
                    PyList_SetItem(mods, j, mod)
                    if mod.is_tracked_for(ModificationCategory_glycosylation):
                        glycosylation_manager[i] = mod
                    mass += mod.mass
            else:
                mods = None

            position = SequencePosition._create(item.amino_acid, mods)
            Py_INCREF(position)
            PyList_SetItem(sequence, i, position)

        self._mass = mass
        if glycan_composition is not None:
            glycosylation_manager.set_aggregate(glycan_composition.clone())

        self._init_termini(
            _make_terminal_group(structure_constants.N_TERM_DEFAULT, n_term),
            _make_terminal_group(structure_constants.C_TERM_DEFAULT, c_term))

    cpdef _init_termini(self, TerminalGroup n_term, TerminalGroup c_term):
        '''This method assumes that the terminal groups are not yet initialized,
        and that their mass contributions are not reflected in :attr:`_mass`
        '''
        self._n_term = n_term
        self._c_term = c_term
        self._mass += n_term.mass + c_term.mass

    cpdef _initialize_fields(self):
        """Initialize all mutable fields to their default, empty values.
        """
        self._mass = 0.0
        self.sequence = []

        self._glycosylation_manager = GlycosylationManager._create(self, None)

        self._n_term = None
        self._c_term = None

        self._total_composition = None
        self._peptide_composition = None

    cpdef _invalidate(self):
        self._total_composition = None
        self._peptide_composition = None

    cdef TerminalGroup get_n_term(self):
        return self._n_term

    cdef void set_n_term(self, _value):
        cdef:
            TerminalGroup value
            double reset_mass, new_mass
        if isinstance(_value, ModificationBase):
            value = self._n_term.modify(_value)
        else:
            value = _value

        self._invalidate()
        reset_mass = 0
        try:
            reset_mass = self._n_term.mass
        except AttributeError:
            pass
        self._n_term = value
        new_mass = 0
        try:
            new_mass = value.mass
        except AttributeError:
            pass
        self._mass += new_mass - reset_mass

    cdef TerminalGroup get_c_term(self):
        return self._c_term

    cdef void set_c_term(self, _value):
        cdef:
            TerminalGroup value
            double reset_mass, new_mass
        if isinstance(_value, ModificationBase):
            value = self._c_term.modify(_value)
        else:
            value = _value
        self._invalidate()
        reset_mass = 0
        try:
            reset_mass = self._c_term.mass
        except AttributeError:
            pass
        self._c_term = value
        new_mass = 0
        try:
            new_mass = value.mass
        except AttributeError:
            pass
        self._mass += new_mass - reset_mass

    def __repr__(self):
        n_term = ""
        if self.n_term is not None:
            n_term = "({0})-".format(self.n_term)
        c_term = ""
        if self.c_term is not None:
            c_term = "-({0})".format(self.c_term)
        aggregate = self._glycosylation_manager.aggregate if self._glycosylation_manager.aggregate is not None else ""
        rep = "{n_term}{sequence}{c_term}{glycan_aggregate}[{_mass}]".format(
            n_term=n_term, c_term=c_term,
            glycan_aggregate=aggregate,
            sequence=self.sequence,
            _mass=self._mass)
        return rep

    @property
    def n_term(self):
        return self.get_n_term()

    @n_term.setter
    def n_term(self, _value):
        self.set_n_term(_value)

    @property
    def c_term(self):
        return self.get_c_term()

    @c_term.setter
    def c_term(self, _value):
        self.set_c_term(_value)

    @property
    def mass(self):
        return self._mass

    @mass.setter
    def mass(self, value):
        self._mass = value

    cdef double get_peptide_backbone_mass(self):
        cdef:
            double total
            size_t i, j, m, n
            SequencePosition position
            ModificationBase mod

        n = self.get_size()
        total = 0
        for i in range(n):
            position = self.get(i)
            total += position.amino_acid.mass
            m = SequencePosition.get_modification_count(position)
            for j in range(m):
                mod = SequencePosition.get_modification(position, j)
                if mod.is_tracked_for(ModificationCategory_glycosylation):
                    continue
                total += mod.mass

        total += self._n_term.mass
        total += self._c_term.mass
        return total

    @property
    def peptide_backbone_mass(self):
        return self.get_peptide_backbone_mass()

    cdef double get_total_mass(self):
        cdef:
            double total
        total = self.get_peptide_backbone_mass()
        gc = self.glycan_composition
        if gc:
            total += PyFloat_AsDouble(gc.mass())
        return total

    @property
    def total_mass(self):
        return self.get_total_mass()

    cpdef CComposition peptide_composition(self):
        if self._peptide_composition is None:
            if self.glycan is None:
                glycan_composition = CComposition()
            else:
                glycan_composition = self.glycan_composition.total_composition()
            self._peptide_base_composition = self.total_composition() - glycan_composition
        return self._peptide_base_composition

    cpdef CComposition total_composition(self):
        if self._total_composition is None:
            self._total_composition = _total_composition(self)
        return self._total_composition

    def __iter__(self):
        return iter(self.sequence)

    def __getitem__(self, index):
        if isinstance(index, slice):
            return self.sequence[index]
        cdef c_index = PyInt_AsLong(index)
        if c_index < 0:
            return self.sequence[index]
        sub = self.get(c_index)
        return sub

    def __setitem__(self, index, value):
        self._invalidate()
        self.sequence[index] = value

    def __len__(self):
        return PyList_Size(self.sequence)

    cdef size_t get_size(self):
        return PyList_Size(self.sequence)

    # equality methods

    def __eq__(self, other):
        if isinstance(other, _PeptideSequenceCore):
            return self.full_structure_equality(<_PeptideSequenceCore>other)
        else:
            return str(self) == str(other)

    cpdef bint base_sequence_equality(self, _PeptideSequenceCore other):
        cdef:
            size_t n, m, i
        n = self.get_size()
        m = other.get_size()
        if n != m:
            return False
        for i in range(n):
            if self.get(i).amino_acid != other.get(i).amino_acid:
                return False
        return True

    cpdef bint modified_sequence_equality(self, _PeptideSequenceCore other):
        cdef:
            size_t n, m, i
            SequencePosition a, b
        n = self.get_size()
        m = other.get_size()
        if n != m:
            return False
        for i in range(n):
            a = self.get(i)
            b = other.get(i)
            if a != b:
                return False
        if self.get_n_term() != other.get_n_term():
            return  False
        if self.get_c_term() != other.get_c_term():
            return False
        return True

    cpdef bint full_structure_equality(self, _PeptideSequenceCore other):
        sequences_equal = self.modified_sequence_equality(other)
        if sequences_equal:
            return self.glycan == other.glycan
        else:
            return False

    def __ne__(self, other):
        return str(self) != str(other)

    def __hash__(self):
        return hash(self.get_sequence())

    cpdef basestring get_sequence(self, bint include_glycan=True, bint include_termini=True, str implicit_n_term=None, str implicit_c_term=None):
        """
        Generate human readable sequence string. Called by :meth:`__str__`

        Parameters
        ----------
        include_glycan: bool
            Whether to include the glycan in the resulting string. Defaults to `True`
        include_termini: bool
            Whether to include the N- and C-termini. Make sure this is `True` if you want non-standard
            termini to be properly propagated.


        Returns
        -------
        str
        """
        if implicit_n_term is None:
            implicit_n_term = structure_constants.N_TERM_DEFAULT
        if implicit_c_term is None:
            implicit_c_term = structure_constants.C_TERM_DEFAULT

        cdef:
            list seq_list,  mod_strs
            size_t i, j, n, m
            SequencePosition position
            ModificationBase mod
            basestring mod_str, n_term, c_term, rep
            bint needs_terminals
            TerminalGroup terminal
            ModificationBase implicit_terminal

        seq_list = []
        n = self.get_size()
        for i in range(n):
            position = self.get(i)
            m = SequencePosition.get_modification_count(position)
            if m > 0:
                if m == 1:
                    mod = SequencePosition.get_modification(position, 0)
                    mod_str = "(%s)" % mod.serialize()
                else:
                    mod_strs = []
                    for j in range(m):
                        mod = SequencePosition.get_modification(position, j)
                        mod_strs.append(mod.serialize())
                    mod_str = '|'.join(mod_strs)
                    mod_str = "(%s)" % mod_str
                seq_list.append(position.amino_acid.symbol + mod_str)
            else:
                seq_list.append(position.amino_acid.symbol)

        rep = ''.join(seq_list)
        if include_termini:
            needs_terminals = False
            # Handle N-terminal
            n_term = ""
            terminal = self.get_n_term()
            implicit_terminal = _get_implicit_terminal(implicit_n_term)
            if terminal is not None and not terminal.equal_to(implicit_terminal):
                n_term = "(%s)-" % terminal.serialize()
                needs_terminals = True

            # Handle C-terminal
            c_term = ""
            terminal = self.get_c_term()
            implicit_terminal = _get_implicit_terminal(implicit_c_term)
            if terminal is not None and not terminal.equal_to(implicit_terminal):
                c_term = "-(%s)" % terminal.serialize()
                needs_terminals = True
            if needs_terminals:
                rep = "".join((n_term, rep, c_term))
        if include_glycan:
            if self._glycosylation_manager.aggregate is not None:
                rep += str(self._glycosylation_manager.aggregate)
        return rep

    def __str__(self):
        return self.get_sequence()

    cpdef clone(self):
        cdef _PeptideSequenceCore inst = self.__class__()
        if self._glycosylation_manager.aggregate is not None:
            glycan = self._glycosylation_manager.aggregate.clone()
            if isinstance(glycan, GlycanCompositionWithOffsetProxy):
                glycan.composition_offset = CComposition._create(None)
            else:
                glycan.composition_offset = CComposition._create(WATER)
        else:
            glycan = None
        inst._init_from_components(
            self.sequence, glycan,
            self.get_n_term().get_modification(),
            self.get_c_term().get_modification())
        return inst


@cython.binding(True)
cpdef add_modification(_PeptideSequenceCore self, position, modification_type):
    '''
    Add a modification by name to a specific residue. If the
    position is the N-term or the C-term, the terminal modification will
    be replaced.

    Parameters
    ----------
    position: int
        The position of the modification to add
    modification_type: str or Modification
        The modification to add
    '''
    cdef:
        long position_
        ModificationBase mod
        SequencePosition pos
    self._invalidate()

    if isinstance(modification_type, Modification):
        mod = modification_type
    else:
        mod = Modification(rule=modification_type)

    if position is SequenceLocation_n_term:
        self.n_term = self.n_term.modify(mod)
    elif position is SequenceLocation_c_term:
        self.c_term = self.c_term.modify(mod)
    else:
        position_ = PyInt_AsLong(position)
        if (position_ == -1) or (position_ >= self.get_size()):
            raise IndexError(
                "Invalid modification position. %s, %s, %s" %
                (position, str(self.sequence), modification_type))
        pos = self.get(position_)
        pos.add_modification(mod)
        self._mass += mod.mass
        if mod.is_tracked_for(ModificationCategory_glycosylation):
            self._glycosylation_manager[position] = mod


from glypy.composition.ccomposition cimport CComposition
from glycopeptidepy._c.structure.base cimport ModificationBase
from glypy.utils.cenum cimport EnumValue

cdef class ModificationRuleBase(ModificationBase):
    cdef:
        public set names
        public list categories

        public basestring preferred_name
        public basestring title
        public basestring common_name
        public basestring unimod_name

        public Py_hash_t _hash

    cpdef basestring serialize(self)




cdef class NeutralLossBase(object):
    cdef:
        public CComposition composition
        public double mass
        public object label


cpdef bint is_tracked_for_glycosylation(self, object category)


cdef class GlycosylationBase(ModificationRuleBase):
    cdef:
        public bint _is_composition
        public bint _is_core
        public EnumValue _glycosylation_type

    cdef bint get_is_composition(self)
    cdef void set_is_composition(self, bint value)

    cdef bint get_is_core(self)
    cdef void set_is_core(self, bint value)

    cdef EnumValue get_glycosylation_type(self)
    cdef void set_glycosylation_type(self, EnumValue glycosylation_type)



import sys
import json
from functools import partial

def add_names(mod_dict, alt_names=None):
    if alt_names is None:
        alt_names = []
    for name in alt_names:
        if name not in mod_dict['alt_names']:
            mod_dict['alt_names'].append(name)

def remove_names(mod_dict, alt_names=None):
    if alt_names is None:
        alt_names = []
    for name in alt_names:
        try:
            mod_dict['alt_names'].remove(name)
        except ValueError as err:
            print(err, name, mod_dict['alt_names'])


patches = [
    {
        "match_key": "full_name",
        "match_value": "Dehydration",
        "changes": [partial(add_names, alt_names=['Dehydrated']), partial(remove_names, alt_names=["Phospho+PL"])]
    }
]


def process(modifications):
    for patch in patches:
        for mod_dict in modifications:
            if mod_dict[patch['match_key']] == patch['match_value']:
                for change in patch['changes']:
                    change(mod_dict)

def main():
    infile = sys.argv[1]
    try:
        outfile = sys.argv[2]
    except IndexError:
        outfile = infile
    with open(infile, 'rt') as fh:
        modification_dicts = json.load(fh)
    process(modification_dicts)
    with open(outfile, 'wt') as fh:
        json.dump(modification_dicts, fh, sort_keys=True, indent=4)

if __name__ == "__main__":
    main()
